/*
 * Copyright (c) 2010 NTNU/KTH author: Oleksandr Bodriagov. All rights reserved. :)

 *  JCpki.java	 15/04/2010

 *  This applet manages RSA public-private key pair and signs data
 *   according to sha-1WithRSAEncryption algorithm. 
 *   
 *   !Cryptographic operations may not be supported in some simulators
 */

package pki;

import javacard.framework.APDU;
import javacard.framework.ISO7816;
import javacard.framework.ISOException;
import javacard.framework.Applet;
import javacard.framework.OwnerPIN;
import javacard.framework.JCSystem;
import javacard.framework.Util;
import javacard.security.*;





public class JCpki extends Applet
{
    final static byte VERIFY_PIN = (byte) 0x20; // INS value for ISO 7816-4 VERIFY
    final static byte UPDATE_PIN = (byte) 0x57; // INS value for update PIN request
    final static byte SIGN = (byte) 0x59; // INS value for the Signature operation
    final static byte GET_PUBLIC_EXPONENT = (byte) 0x54; // INS value
    final static byte GET_PUBLIC_MODULUS = (byte) 0x55; // INS value   
    final static byte UPDATE_KEYS = (byte) 0x46; // INS value for ISO 7816-4 GENERATE ASYMMETRIC KEY PAIR
    final static byte [] SW_NO_ERROR = {(byte)0x90,(byte)0x00};

    final static byte TryLimit = (byte) 4;// number of tries for PIN
    final static byte MaxPINSize = (byte) 8;
    
    private KeyPair rsaPair;
    private OwnerPIN appletPIN;
    private RSAPrivateKey rsaPrivateKey; 
    private RSAPublicKey rsaPublicKey;
    private Signature mSignature;
    
    private short dataLength; 
    private byte[] signedbuffer; //array for storing signed data
    private short signatureLength;
    
     
    // Constructor. Only this class's install method should create the applet object.
    protected JCpki(byte[] installParameters, short offset, byte length)
    {
    	byte [] initialPIN = new byte[4];
        short pinOffset = offset;        
        
        Util.arrayCopy(installParameters, (short)(pinOffset), initialPIN,(short)0 , (short)4);
        appletPIN = new OwnerPIN(TryLimit, MaxPINSize);
        appletPIN.update(initialPIN, (short) 0,(byte) 4);
        
        
        //generate public-private key pair
        //rsaPair = new KeyPair( KeyPair.ALG_RSA, KeyBuilder.LENGTH_RSA_2048 );
        rsaPair = new KeyPair( KeyPair.ALG_RSA, (short)512 );
        rsaPair.genKeyPair();
        rsaPublicKey = (RSAPublicKey) rsaPair.getPublic();
        rsaPrivateKey = (RSAPrivateKey) rsaPair.getPrivate();

        //initialize signature
        mSignature = Signature.getInstance(Signature.ALG_RSA_SHA_PKCS1, false);
        mSignature.init(rsaPrivateKey, Signature.MODE_SIGN);
        
        signedbuffer = JCSystem.makeTransientByteArray((short)128, JCSystem.MEMORY_TYPE_TRANSIENT_RESET);//cleared when applet is deselected
       
        register();
    }


    // Installs applet. installParameters: contains installation parameters ( PIN, )
    public static void install(byte[] bArray, short bOffset, byte bLength)
    {
        new JCpki(bArray, bOffset, bLength);
    }



    // Processes incoming APDUs
    public void process(APDU apdu)
    {
        byte[] buffer = apdu.getBuffer(); //gets a pointer to the APDU buffer

	    // check if it is SELECT APDU command
        if ( buffer[ISO7816.OFFSET_CLA] == 0 && buffer[ISO7816.OFFSET_INS] == (byte) (0xA4) )
        return;

        switch (buffer[ISO7816.OFFSET_INS])
        {
             case VERIFY_PIN:
                    verifyPIN(apdu);
                    break;
             case SIGN:
                    signData(apdu);
                    break;
             case UPDATE_PIN:
                    updatePIN(apdu);
                    break;
             case GET_PUBLIC_EXPONENT:
            	    sendExponent(apdu);
            	    break;
             case GET_PUBLIC_MODULUS:
            	    sendModulus(apdu);
         	        break;
             case UPDATE_KEYS:
            	    updateKeys(apdu);
            	    break;
             default:
                    ISOException.throwIt(ISO7816.SW_INS_NOT_SUPPORTED);
        }
    }



    // user authentication via PIN check
    private void verifyPIN(APDU apdu)
    {
        byte[] buffer = apdu.getBuffer();
        byte dataLength = buffer[ISO7816.OFFSET_LC]; // number of bytes in the data field of APDU
        if (dataLength > 8 || dataLength < 1)
            ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
        apdu.setIncomingAndReceive(); //data is read into apdu buffer

        if (appletPIN.check(buffer, ISO7816.OFFSET_CDATA, dataLength ) != true)
        	ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
        else
        {
        	apdu.setOutgoing();
            apdu.setOutgoingLength((short) SW_NO_ERROR.length);
            apdu.sendBytesLong(SW_NO_ERROR, (short) 0, (short)SW_NO_ERROR.length);
        }
    }
    
    private void updatePIN(APDU apdu)
    {
        if(appletPIN.isValidated() != true)
            ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
        appletPIN.reset();//resets the validated flag and resets the PIN try counter to the value
        
        byte[] buffer = apdu.getBuffer();
        byte dataLength = buffer[ISO7816.OFFSET_LC]; // number of bytes in the data field of APDU
        apdu.setIncomingAndReceive(); //data is read into apdu buffer
        
        if (dataLength > 8 || dataLength < 4)
            ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
        appletPIN.update(buffer, ISO7816.OFFSET_CDATA, dataLength);
        appletPIN.resetAndUnblock();
        
    	apdu.setOutgoing();
        apdu.setOutgoingLength((short) SW_NO_ERROR.length);
        apdu.sendBytesLong(SW_NO_ERROR, (short) 0, (short)SW_NO_ERROR.length);
	}


    private void sendExponent(APDU apdu)
    {
    	if(appletPIN.isValidated() != true)
            ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
        appletPIN.reset();//resets the validated flag and resets the PIN try counter to the value
        try
        {
        	dataLength = rsaPublicKey.getExponent(signedbuffer, (short) 0);
        }
        catch (CryptoException e)
        {
        	ISOException.throwIt(e.getReason());
        	return;
        }
        apdu.setOutgoing();
        apdu.setOutgoingLength((short) dataLength);
        apdu.sendBytesLong(signedbuffer, (short) 0, (short) dataLength);        	
    }
    
    
    private void sendModulus (APDU apdu)
    {
    	if(appletPIN.isValidated() != true)
            ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
        appletPIN.reset();//resets the validated flag and resets the PIN try counter to the value
        
        try
        {
        	dataLength = rsaPublicKey.getModulus(signedbuffer, (short)0);
        }
        catch (CryptoException e)
        {
        	ISOException.throwIt(e.getReason());
        	return;
        }
        apdu.setOutgoing();
        apdu.setOutgoingLength((short) dataLength);
        apdu.sendBytesLong(signedbuffer, (short) 0, (short) dataLength);
    }
    
    
    private void updateKeys(APDU apdu)
    {
    	if(appletPIN.isValidated() != true)
            ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
        appletPIN.reset();//resets the validated flag and resets the PIN try counter to the value
        
        try
        {
	        rsaPair.genKeyPair();
	        rsaPublicKey = (RSAPublicKey) rsaPair.getPublic();
	        rsaPrivateKey = (RSAPrivateKey) rsaPair.getPrivate();
        }
        catch (CryptoException e)
        {
        	ISOException.throwIt(e.getReason());
        	return;
        }
                
    	apdu.setOutgoing();
        apdu.setOutgoingLength((short) SW_NO_ERROR.length);
        apdu.sendBytesLong(SW_NO_ERROR, (short) 0, (short)SW_NO_ERROR.length);    	
    }

    
    private void signData(APDU apdu)
    {
        if(appletPIN.isValidated() != true)
            ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
        appletPIN.reset();//resets the validated flag and resets the PIN try counter to the value

        byte[] buffer = apdu.getBuffer();
        dataLength = buffer[ISO7816.OFFSET_LC]; // number of bytes in the data field of APDU
        apdu.setIncomingAndReceive(); //data is read into the Apdu buffer

        try
        {
        	signatureLength = mSignature.sign(buffer, (short)ISO7816.OFFSET_CDATA, dataLength, signedbuffer, (short) 0);
        }
        catch (CryptoException e)
        {
        	ISOException.throwIt(e.getReason());
        	return;
        }
        apdu.setOutgoing();
        apdu.setOutgoingLength((short) signatureLength);
        apdu.sendBytesLong(signedbuffer, (short) 0, signatureLength);
    }




}//end class JCpki