/*
 * Copyright (c) 2010 NTNU/KTH; author: Oleksandr Bodriagov. All rights reserved. :)
 
 * Class CardCommunicator: handles communication with the Java Card applet
 */


package pkiphone;

import javax.microedition.apdu.APDUConnection;


public class CardCommunicator
{
    private APDUConnection cardConnection;

    final static byte [] SW_NO_ERROR = {(byte)0x90,(byte)0x00};
    final static byte INS_VERIFY_PIN = (byte) 0x20; // INS value for ISO 7816-4 VERIFY
    final static byte INS_CHANGE_PIN = (byte) 0x57; // INS value for Update PIN command


    public CardCommunicator (APDUConnection connection)
    {
        cardConnection = connection;
    }

    
    //sends data to the card for signing
    public byte[] sign( byte [] data)
    {
        if (cardConnection == null)
            return null;
        try
        {
            byte[] response = cardConnection.exchangeAPDU(makeSignatureApdu( data));

            if (response.length > 2)//received some data, not the error notification
                return response;
        }
        catch (Exception e)
        {
            return null;
        }
        return null;
    }

    // sends a command to the card to update the key pair
    public boolean updateKeys()
    {
        if (cardConnection == null)
            return false;

        byte [] updateKeysApdu = {(byte)0x80,(byte)0x46, (byte)0x00,(byte)0x00, (byte)0x00};
        try
        {
            byte[] response = cardConnection.exchangeAPDU(updateKeysApdu);

            if (response[0] == SW_NO_ERROR[0] && response[1] == SW_NO_ERROR[1])
                return true;
            else
                return false;
        }
        catch (Exception e)
        {
            return false;
        }
    }


    //sends PIN to the Java Card applet for verification
    public boolean verifyPIN (String pinCode)
    {
        if (cardConnection == null)
        {
           // if (pinCode.compareTo("1215") == 0)
             //   return true;

            return false;
        }
        
        try
        {
            byte[] response = cardConnection.exchangeAPDU(makePinApdu(INS_VERIFY_PIN, pinCode));

            if (response[0] == SW_NO_ERROR[0] && response[1] == SW_NO_ERROR[1])
                return true;
            else
                return false;
        }
        catch (Exception e)
        {
            return false;            
        }
        
    }

    //sends new String PIN to the Java Card applet for verification
    public boolean updatePIN (String newPIN)
    {
        if (cardConnection == null)
            return false;
        
        try
        {
            byte[] response = cardConnection.exchangeAPDU(makePinApdu(INS_CHANGE_PIN, newPIN ));

            if (response[0] == SW_NO_ERROR[0] && response[1] == SW_NO_ERROR[1])
                return true;
            else
                return false;
        }
        catch (Exception e)
        {
            return false;
        }
    }


    //sends new byte [] PIN to the Java Card applet for verification
    public boolean updatePIN (byte [] newPIN)
    {
        if (cardConnection == null)
            return false;

        try
        {
            byte[] response = cardConnection.exchangeAPDU(makePinApdu(INS_CHANGE_PIN, newPIN ));

            if (response[0] == SW_NO_ERROR[0] && response[1] == SW_NO_ERROR[1])
                return true;
            else
                return false;
        }
        catch (Exception e)
        {
            return false;
        }
    }



    //creates envelope that contains entered PIN from a String
    private byte[] makePinApdu(byte command, String pin)
    {
        if (pin == null)
            return null;
        
        int pinlength = pin.length();
        byte [] buffer = pin.getBytes();
        byte [] pinEnvelope = new byte [6 + pinlength];
        pinEnvelope [0] = (byte)0x80;//APDU CLA

        switch(command)
        {
            case INS_VERIFY_PIN:
                pinEnvelope [1] = (byte)INS_VERIFY_PIN;//APDU INS
                break;
            case INS_CHANGE_PIN:
                pinEnvelope [1] = (byte)INS_CHANGE_PIN;//APDU INS
                break;
            default:
                return null;
        }

        pinEnvelope [2] = (byte)0;//APDU P1
        pinEnvelope [3] = (byte)0;//APDU P2
        pinEnvelope [4] = (byte)pinlength;//APDU Lc
        for (int i=0; i < pinlength; i++)
            pinEnvelope [5+i] = buffer[i];

        pinEnvelope [5 + pinlength] = (byte)0;
        return pinEnvelope;
     }


    //creates envelope that contains entered PIN from a byte array
    private byte[] makePinApdu(byte command, byte [] pin)
    {
        if (pin == null)
            return null;

        int pinlength = pin.length;
        byte [] pinEnvelope = new byte [6 + pinlength];
        pinEnvelope [0] = (byte)0x80;//APDU CLA

        switch(command)
        {
            case INS_VERIFY_PIN:
                pinEnvelope [1] = (byte)INS_VERIFY_PIN;//APDU INS
                break;
            case INS_CHANGE_PIN:
                pinEnvelope [1] = (byte)INS_CHANGE_PIN;//APDU INS
                break;
            default:
                return null;
        }

        pinEnvelope [2] = (byte)0;//APDU P1
        pinEnvelope [3] = (byte)0;//APDU P2
        pinEnvelope [4] = (byte)pinlength;//APDU Lc
        for (int i=0; i < pinlength; i++)
            pinEnvelope [5+i] = pin[i];

        pinEnvelope [5 + pinlength] = (byte)0;
        return pinEnvelope;
     }


     //creates envelope that contains data for signing
     private byte [] makeSignatureApdu(byte [] data)
     {
         byte [] apdu = new byte [6 + data.length];
         apdu [0] = (byte)0x80;//APDU CLA
         apdu [1] = (byte)0x59;//APDU INS_SIGN
         apdu [2] = (byte)0;//APDU P1
         apdu [3] = (byte)0;//APDU P2
         apdu [4] = (byte)data.length;//APDU LC - number of bytes present in the data field

         for (int i=0; i < data.length; i++)
             apdu [5+i] = data[i];//APDU Data

         apdu [5 + data.length] = (byte)0;//APDU Le

         return apdu;
     }

}
