/*++

Copyright (c) Microsoft Corporation.  All rights reserved.

    THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
    KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
    PURPOSE.

Abstract:
    Utility for installing and uninstalling BthScrdSrv - bluetooth smart card reader server.

    bthsrvinst /i (installs bth service) 
    bthsrvinst /u (uninstalls bth service)
--*/

#include <windows.h>
#include <bthsdpdef.h>
#include <bthdef.h>
#include <BlueToothApis.h>
#include <stdio.h>
#include <strsafe.h>
#include <initguid.h>
#include "public.h"

#define ASSERT(b)

#define SWITCH_ARG_LEN 2

DWORD ReadSwitch( __in LPCSTR SwitchArg, __out char * Switch);

BOOL AdjustProcessPrivileges( );

DWORD SetBthServiceInfo( BOOLEAN bEnabled );




void __cdecl main( __in UINT nArgs, __in_ecount(nArgs) LPCSTR Args[]  )
{
    if (nArgs < 2)
    {
        printf("Usage: bthsrvinst [/i | /u]\n");
		return;
    }

    char Switch;

    if (ERROR_SUCCESS != ReadSwitch(Args[1], &Switch))
        return;

    if (FALSE == AdjustProcessPrivileges())
		return;

    switch(Switch)
    {
        case 'i':
            SetBthServiceInfo(TRUE);
            break;
        case 'u':
            SetBthServiceInfo(FALSE);
            break;
        default:
			printf("Usage: bthsrvinst [/i | /u]\n");
			return;
    }    
}



DWORD SetBthServiceInfo( BOOLEAN bEnabled )
{
    DWORD err = ERROR_SUCCESS;
    BLUETOOTH_LOCAL_SERVICE_INFO SvcInfo = {0};//structure that describes the local service to set
    SvcInfo.Enabled = bEnabled;

	//BthScrdSvcName defined in public.h
    if (FAILED(StringCbCopyW(SvcInfo.szName, sizeof(SvcInfo.szName), BthScrdSvcName)))
    {
        printf("Copying svc name failed\n");
        return err;
    }

	// BluetoothSetLocalServiceInfo sets local service information for a specific Bluetooth radio
	// Since a handle of the Bluetooth radio device = NULL, it searches for the first  available local Bluetooth radio
	err = BluetoothSetLocalServiceInfo(
		NULL, //searches for the first  available local Bluetooth radio
        &BTHSCRD_SVC_GUID,
        0,
        &SvcInfo
        );
    if (err != ERROR_SUCCESS)
        printf("BluetoothSetLocalServiceInfo failed, err = %d\n", err);      

    return err;    
}



DWORD ReadSwitch( __in LPCSTR SwitchArg,  __out char * Switch )
{
    DWORD retval = ERROR_INVALID_PARAMETER;
    
    size_t len = 0;
    if (FAILED(StringCchLength(SwitchArg, SWITCH_ARG_LEN+1, &len)) ||
        len != SWITCH_ARG_LEN)
    {
        printf("Invalid arg %s, expected arg length: %d, actual length: %d\n", SwitchArg, SWITCH_ARG_LEN, len);
        return retval;
    }

    if (SwitchArg[0] != L'/' && SwitchArg[0] != L'-')
    {
        printf("Invalid first character in arg %s\n", SwitchArg);
        return retval;        
    }

    retval = ERROR_SUCCESS;
    *Switch = SwitchArg[1];

    return retval;
}

BOOL AdjustProcessPrivileges()
{
    HANDLE procToken;
    LUID luid;
    TOKEN_PRIVILEGES tp;
    BOOL bRetVal;
    DWORD err;

    bRetVal = OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &procToken);
    if (!bRetVal)
    {
        err = GetLastError();
        printf("OpenProcessToken failed, err = %d\n", err);
        goto exit;
    }
        
    bRetVal = LookupPrivilegeValue(NULL, SE_LOAD_DRIVER_NAME, &luid);
    if (!bRetVal)
    {
        err = GetLastError();
        printf("LookupPrivilegeValue failed, err = %d\n", err);
        goto exit1;
    }

    tp.PrivilegeCount = 1;
    tp.Privileges[0].Luid = luid;
    tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

    //
    // AdjustTokenPrivileges can succeed even when privileges are not adjusted.
    // In such case GetLastError returns ERROR_NOT_ALL_ASSIGNED.
    //
    // Hence we check for GetLastError in both success and failure case.
    //

    (void) AdjustTokenPrivileges(procToken, FALSE, &tp, sizeof(TOKEN_PRIVILEGES), (PTOKEN_PRIVILEGES) NULL, (PDWORD)NULL);
    err = GetLastError();

    if (err != ERROR_SUCCESS)
    {
        bRetVal = FALSE;
        printf("AdjustTokenPrivileges failed, err = %d\n", err);
        goto exit1;
    }

exit1:
    CloseHandle(procToken);
exit:
    return bRetVal;
}

