/*
 *  brbservercontext.c

 *  contains brb functions
 */

#include "brbservercontext.h"

__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoSharedSendBrbAsync(
    __in WDFIOTARGET IoTarget,
    __in WDFREQUEST Request,
    __in PBRB Brb,
    __in size_t BrbSize,
    __in PFN_WDF_REQUEST_COMPLETION_ROUTINE ComplRoutine,
    __in_opt WDFCONTEXT Context
    )
/*++
BthEchoSharedSendBrbAsync: This routine formats a request with brb and sends it asynchronously

Arguments:
    IoTarget - Target to send the brb to
    Request - request object to be formatted with brb                
    Brb - Brb to be sent
    BrbSize - size of the Brb data structure
    ComplRoutine - WDF completion routine for the request. This must be specified because
	               we are formatting the request and hence not using SEND_AND_FORGET flag
    Context - (optional) context to be passed in to the completion routine

Return Value:
    Success implies that request was sent correctly and completion routine will be called
    for it, failure implies it was not sent and caller should complete the request

Notes:
    This routine does not call WdfRequestReuse on the Request passed in.
    Caller must do so before passing in the request, if it is reusing the request.

    This routine does not complete the request in case of failure.
    Caller must complete the request in case of failure.

--*/
{
    NTSTATUS status;
    WDF_OBJECT_ATTRIBUTES attributes;
    WDFMEMORY memoryArg1;

	KdPrint(("SCRD! BthEchoSharedSendBrbAsync: Enter\n"));

    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.ParentObject = Request;

    status = WdfMemoryCreatePreallocated(
        &attributes,
        Brb,
        BrbSize,
        &memoryArg1
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("Creating preallocted memory for Brb 0x%p failed, Request to be formatted 0x%p, Status code %!STATUS!\n",
			Brb, Request, status));

        return status;
    }

    status = WdfIoTargetFormatRequestForInternalIoctlOthers(
        IoTarget,
        Request,
        IOCTL_INTERNAL_BTH_SUBMIT_BRB,
        memoryArg1,
        NULL, //OtherArg1Offset
        NULL, //OtherArg2
        NULL, //OtherArg2Offset
        NULL, //OtherArg4
        NULL  //OtherArg4Offset
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("Formatting request 0x%p with Brb 0x%p failed, Status code %x\n",
			Request, Brb, status));

        return status;
    }

    //
    // Set a CompletionRoutine callback function.
    //
    WdfRequestSetCompletionRoutine(
        Request,
        ComplRoutine,
        Context
        );

    if (FALSE == WdfRequestSend(
        Request,
        IoTarget,
        NULL
        ))
    {
        status = WdfRequestGetStatus(Request);

		KdPrint( ("Request send failed for request 0x%p, Brb 0x%p, Status code %s\n",
			Request, Brb, status));

    }

	KdPrint(("SCRD! BthEchoSharedSendBrbAsync: Exit\n"));

    return status;
}




__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
//FORCEINLINE
BthEchoSharedSendBrbSynchronously(
    __in WDFIOTARGET IoTarget,
    __in WDFREQUEST Request,
    __in PBRB Brb,
    __in ULONG BrbSize
    )
/*++
Description:
    This routine formats a request with brb and sends it synchronously

Arguments:
    IoTarget - Target to send the brb to
    Request - request object to be formatted with brb                
    Brb - Brb to be sent
    BrbSize - size of the Brb data structure

Notes:
    This routine does calls WdfRequestReuse on the Request passed in.
    Caller need not do so before passing in the request.

    This routine does not complete the request in case of failure.
    Caller must complete the request in case of failure.
--*/
{
    NTSTATUS status;
    WDF_REQUEST_REUSE_PARAMS reuseParams;
    WDF_MEMORY_DESCRIPTOR OtherArg1Desc;

	KdPrint(("SCRD! BthEchoSharedSendBrbSynchronously: Enter\n"));

    WDF_REQUEST_REUSE_PARAMS_INIT(
        &reuseParams,
        WDF_REQUEST_REUSE_NO_FLAGS, 
        STATUS_NOT_SUPPORTED
        );

	//reinitializes a framework request object so that it can be reused
    status = WdfRequestReuse(Request, &reuseParams);
    if (!NT_SUCCESS(status))
    {
        return status;
    }

    WDF_MEMORY_DESCRIPTOR_INIT_BUFFER(
                        &OtherArg1Desc,
                        Brb,
                        BrbSize
                        );
    //builds a non-standard internal device control request and sends it synchronously to an I/O target
    status = WdfIoTargetSendInternalIoctlOthersSynchronously(
        IoTarget,
        Request,
        IOCTL_INTERNAL_BTH_SUBMIT_BRB,
        &OtherArg1Desc,
        NULL, //OtherArg2
        NULL, //OtherArg4
        NULL, //RequestOptions
        NULL  //BytesReturned
        );

	KdPrint(("SCRD! BthEchoSharedSendBrbSynchronously: Exit\n"));
    return status;
}