/*
 * Copyright (c) 2010 NTNU/KTH author: Oleksandr Bodriagov. All rights reserved. :)

 *  brbservercontext.c - main file

 *  contains initialization, registration, PnP functions, and I/O request handling
 */


#include "bthpcsc.h"
#include "serverblabla.h"
#include "SCardExtCallback.h"
#include "sdp.h"

#include "PcscLog.h"
#include "BthCrVersion.h"



//The alloc_text pragma names the code section where the specified function definitions are to reside.
//The routines in the INIT section are not pageable and are discarded as soon as the driver returns from its DriverEntry
#pragma alloc_text(INIT, DriverEntry)
#pragma alloc_text(PAGEABLE, PcscEvtDeviceAdd)//pageable code section
#pragma alloc_text(PAGEABLE, PcscRegisterWithSmcLib)
#pragma alloc_text (PAGE, BthScrdEvtDeviceSelfManagedIoCleanup)




/*++
DriverEntry:
   entry function of the driver. setup the callbacks for the OS and try to
   initialize a device object for every device in the system

Arguments:
   DriverObject   context of the driver
   RegistryPath   path to the registry entry for the driver

Return Value:
   STATUS_SUCCESS
   STATUS_UNSUCCESSFUL
--*/

NTSTATUS
DriverEntry(
           PDRIVER_OBJECT DriverObject,
           PUNICODE_STRING   RegistryPath //path to the registry entry for the driver
           )
{

    NTSTATUS status = STATUS_SUCCESS;
    WDF_DRIVER_CONFIG   config;

	KdPrint(("\n\nSCRD! DriverEntry: Enter\n"));

	//to initialize the driver object configuration structure with a pointer to the driver's EvtDriverDeviceAdd callback
    WDF_DRIVER_CONFIG_INIT(&config, PcscEvtDeviceAdd); 


    // Create a framework driver object to represent our driver.
    status = WdfDriverCreate(DriverObject,
                        RegistryPath,
                        WDF_NO_OBJECT_ATTRIBUTES, // Driver Attributes
                        &config,          // Driver Config Info
                        WDF_NO_HANDLE
                        );

    if (!NT_SUCCESS(status)) {
        KdPrint( ("WdfDriverCreate failed with status 0x%x\n", status));
    }

	KdPrint(("SCRD! DriverEntry: Exit\n"));

    return status;
}



/*++
PcscEvtDeviceAdd:
    PscrEvtDeviceAdd is called by the framework in response to AddDevice
    call from the PnP manager. We create and initialize a WDF device object to
    represent a new instance of the device.

Arguments:
    Driver - Handle to a framework driver object created in DriverEntry
    DeviceInit - Pointer to a framework-allocated WDFDEVICE_INIT structure.
--*/
NTSTATUS
PcscEvtDeviceAdd(
    IN WDFDRIVER       Driver, //Handle to a framework driver object created in DriverEntry
    IN PWDFDEVICE_INIT DeviceInit
    )
{
	NTSTATUS status;
	WDF_PNPPOWER_EVENT_CALLBACKS pnpPowerCallbacks; //WDF_PNPPOWER_EVENT_CALLBACKS structure
	WDF_OBJECT_ATTRIBUTES attributes;
	WDFDEVICE device;
	PBTH_DEVICE_CONTEXT pDevContext;
	WDF_IO_QUEUE_CONFIG ioQueueConfig;
	WDF_FILEOBJECT_CONFIG fileobjectConfig;

	UNREFERENCED_PARAMETER(Driver);

	PAGED_CODE();

	KdPrint(("BTH SCRD! PcscAddDevice: Enter\n"));

    WdfDeviceInitSetDeviceType(DeviceInit, FILE_DEVICE_SMARTCARD); //sets the device type for a specified device

    WdfDeviceInitSetExclusive(DeviceInit, TRUE);

    WdfDeviceInitSetIoType(DeviceInit, WdfDeviceIoBuffered);

    WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&pnpPowerCallbacks);

    
	// EvtDeviceSelfManagedIoCleanup callback function is called by the framework
	// after the specified device has been removed from the system
	pnpPowerCallbacks.EvtDeviceSelfManagedIoInit = BthScrdEvtDeviceSelfManagedIoInit;//initializes and starts the device's self-managed I/O operations
    pnpPowerCallbacks.EvtDeviceSelfManagedIoCleanup = BthScrdEvtDeviceSelfManagedIoCleanup;

    WdfDeviceInitSetPnpPowerEventCallbacks(DeviceInit, &pnpPowerCallbacks);


	WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes, BTH_DEVICE_CONTEXT);
	attributes.EvtCleanupCallback = PcscEvtDeviceContextCleanup;

	//The framework calls the cleanup callback before it deletes the object,
	//the object's parent, or any of the object's more distant ancestors.

	/*The driver passes the attributes structure and the WDFDEVICE_INIT structure to WdfDeviceCreate
	to create the device object. WdfDeviceCreate creates a WDFDEVICE object, attaches the device
	object to the Plug and Play device stack, and returns a handle to the object.*/
	
	status = WdfDeviceCreate(&DeviceInit, &attributes, &device);
	if (!NT_SUCCESS(status))
	{
		//SmartcardLogError routine writes an error message to the event log
        SmartcardLogError(
                 WdfDriverWdmGetDriverObject(WdfGetDriver()),
                 PSCR_INSUFFICIENT_RESOURCES,
                 NULL,
                 0
                 );
		return status;
	}
	pDevContext = GetServerDeviceContext(device);
	pDevContext->Device = device;
	pDevContext->IoTarget = WdfDeviceGetIoTarget(device);



    ///////////////////////////////////////////////////////////////
	//         Bluetooth part server context initialization
	///////////////////////////////////////////////////////////////

	//Initialize request object
    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.ParentObject = device;

    status = WdfRequestCreate(
        &attributes,
        pDevContext->IoTarget,
        &pDevContext->Request
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("Failed to pre-allocate request in device context, Status code %!STATUS!\n", status));

        return status;        
    }

	WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.ParentObject = device;

    status = WdfSpinLockCreate(
                               &attributes,
                               &pDevContext->ConnectionListLock
                               );
    if (!NT_SUCCESS(status))
    {
		 KdPrint( ("Initialization of context failed with Status code %!STATUS!\n", status));
        return status;                
    }

	//initializes a LIST_ENTRY structure that represents the head of a doubly linked list
    InitializeListHead(&pDevContext->ConnectionList);

	//-end of BthEchoSampleServerContextInit------------------------------------------------------


    // Query for interfaces and pre-allocate BRBs

    /* WdfFdoQueryForInterface method obtains access to another driver's GUID-identified
	interface. Profile drivers query for the BTHDDI_PROFILE_DRIVER_INTERFACE to obtain 
	pointers to functions that allow them to create, allocate, reuse, and free BRBs.*/
	status = WdfFdoQueryForInterface(
			device,
			&GUID_BTHDDI_PROFILE_DRIVER_INTERFACE, 
			(PINTERFACE) (&pDevContext->ProfileDrvInterface), //A pointer to a driver-allocated structure that receives the requested interface
			sizeof(pDevContext->ProfileDrvInterface), 
			BTHDDI_PROFILE_DRIVER_INTERFACE_VERSION_FOR_QI,//The version number of the requested interface
			NULL
			);  
	if (!NT_SUCCESS(status))
	{
		KdPrint(("QueryInterface failed for Interface profile driver interface, version %d, Status code %s\n",
                BTHDDI_PROFILE_DRIVER_INTERFACE_VERSION_FOR_QI, status));

		return status;
	}   

	///////////////////////////////////////////////////////////////
	//         Bluetooth part end
	///////////////////////////////////////////////////////////////



	InterlockedIncrement((PLONG)&pDevContext->DeviceInstanceNo);


	// Add an extra stack location to the IRP dispatched to our device so
	// that we can use that to workaround the problem of smclib	completing
	// ioctl IRPs.

	WdfDeviceWdmGetDeviceObject(pDevContext->Device)->StackSize++;//Returns a pointer to the WDM DEVICE_OBJECT that is associated with a specified framework device object



    // Tell the Framework that this device will need an interface
    status = WdfDeviceCreateDeviceInterface(
                 device,
                 &SmartCardReaderGuid,
                 NULL // ReferenceString
             );

    if (!NT_SUCCESS (status)) {
        SmartcardLogError(
                         WdfDriverWdmGetDriverObject(WdfGetDriver()),
                         PSCR_INSUFFICIENT_RESOURCES,
                         NULL,
                         0
                         );

        return status;
    }

    //
    // Create a parallel queue for dispatching ioctl requests.
	WDF_IO_QUEUE_CONFIG_INIT(
		&ioQueueConfig,
		WdfIoQueueDispatchParallel
		);

	ioQueueConfig.EvtIoDeviceControl = PscrEvtIoDeviceControl;//Sets the event callbacks for the queue

    status = WdfIoQueueCreate (
                   device,
                   &ioQueueConfig,
                   WDF_NO_OBJECT_ATTRIBUTES,
                   &pDevContext->IoctlQueue
                   );
    if(!NT_SUCCESS (status)){
		return status;
    }

	//Specifies which I/O requests KMDF should direct to the queue
    status = WdfDeviceConfigureRequestDispatching(
		device,
		pDevContext->IoctlQueue,
		WdfRequestTypeDeviceControl);

    if(!NT_SUCCESS (status)){
        return status;
    }


	 //
    // Manual queue for parking smartcard notification request.
    //
    WDF_IO_QUEUE_CONFIG_INIT(
        &ioQueueConfig,
        WdfIoQueueDispatchManual
        );

    //
    // Since this request dispatch from this queue doesn't touch the 
    // hardware, it doesn't have to be power managed.
    //
    ioQueueConfig.PowerManaged = WdfFalse;

    //
    // Smartcard library keeps track of last outstanding notification request provided
	// to the client driver. When the client driver completes the request, it has to clear
	// the field in the data-structure maintained by the library. Since we keep the request
    // in a queue, we need a notification from the framework when the 
    // request is cancelled so that we can clear the field => PscrEvtIoCanceledOnQueue callback.
    //

	//The system can request that a driver stop processing an I/O request at any time
	// for many reasons. If the request has been delivered but the driver forwards it
	// to a different queue, the framework automatically cancels the request without notifying the driver
    ioQueueConfig.EvtIoCanceledOnQueue = PscrEvtIoCanceledOnQueue;

    status = WdfIoQueueCreate (
                   device,
                   &ioQueueConfig,
                   WDF_NO_OBJECT_ATTRIBUTES,
                   &pDevContext->NotificationQueue
                   );

    if(!NT_SUCCESS (status)){
        return status;
    }

    /*
    // Create WDFINTERRUPT object. Framework will take care of parsing
    // the interrupt resource and connecting and disconnecting the ISR
    // at the appropriate PNP/POWER state changes.
    //

    WDF_INTERRUPT_CONFIG_INIT(&interruptConfig,
                              PscrEvtInterruptServiceRoutine,
                              PscrEvtInterruptDpc);

    status = WdfInterruptCreate(device,
                        &interruptConfig,
                        WDF_NO_OBJECT_ATTRIBUTES,
                        &pDevContext->Interrupt);
    if (!NT_SUCCESS (status)) {
        return status;
    }
	*/

	// Register with smartcard library.
    status = PcscRegisterWithSmcLib(pDevContext);
    if (!NT_SUCCESS (status)) {
        return status;
    }

	// initialize synchronization event: wait for client's response before 
	// returning transmit status to smclib
	KeInitializeEvent(
		&pDevContext->ReceivedResponseEvt,
		SynchronizationEvent,
		FALSE
		);

    return status;
}





NTSTATUS
PcscRegisterWithSmcLib(
    PBTH_DEVICE_CONTEXT pDevContext
    )
{
    PREADER_EXTENSION ReaderExtension;
	//SMARTCARD_EXTENSION structure is used by both the smart card reader driver
	//and the smart card driver library to access all other smart card data structures
    PSMARTCARD_EXTENSION SmartCardExtension;
    UNICODE_STRING vendorNameU, ifdTypeU;
    ANSI_STRING vendorNameA, ifdTypeA;
    WDFSTRING vendorNameStr, ifdTypeStr;
    NTSTATUS status;

    PAGED_CODE();

	vendorNameStr = NULL;
    ifdTypeStr = NULL;

    RtlZeroMemory(&vendorNameA, sizeof(vendorNameA));
    RtlZeroMemory(&ifdTypeA, sizeof(ifdTypeA));

	SmartCardExtension = &pDevContext->SmartCardExtension;

	// allocate the reader extension
    ReaderExtension = ExAllocatePoolWithTag(
                                    NonPagedPool,
                                    sizeof( READER_EXTENSION ),
                                    SMARTCARD_POOL_TAG
                                    );

    if ( ReaderExtension == NULL ) {
        SmartcardLogError(
                         WdfDriverWdmGetDriverObject(WdfGetDriver()),
                         PSCR_INSUFFICIENT_RESOURCES,
                         NULL,
                         0
                         );
        status = STATUS_INSUFFICIENT_RESOURCES;
        return status;
    }

    RtlZeroMemory( ReaderExtension, sizeof( READER_EXTENSION ));

    SmartCardExtension->ReaderExtension = ReaderExtension;

	ReaderExtension->CompletionRoutine  = CBUpdateCardState;
    ReaderExtension->TrackingContext    = SmartCardExtension;



	   //   setup smartcard extension - callback's
    SmartCardExtension->ReaderFunction[RDF_CARD_POWER] = CBCardPower;
    SmartCardExtension->ReaderFunction[RDF_TRANSMIT] = CBTransmit;
    SmartCardExtension->ReaderFunction[RDF_CARD_TRACKING] = CBCardTracking;
    SmartCardExtension->ReaderFunction[RDF_SET_PROTOCOL] = CBSetProtocol;
    //SmartCardExtension->ReaderFunction[RDF_IOCTL_VENDOR] = PscrGenericIOCTL;

	
	// setup smartcard extension - vendor attribute
    RtlCopyMemory(
                 SmartCardExtension->VendorAttr.VendorName.Buffer,//destination
                 PSCR_VENDOR_NAME,//source
                 sizeof( PSCR_VENDOR_NAME )
                 );
    SmartCardExtension->VendorAttr.VendorName.Length = sizeof( PSCR_VENDOR_NAME );

    RtlCopyMemory(
                 SmartCardExtension->VendorAttr.IfdType.Buffer,
                 PSCR_IFD_TYPE,
                 sizeof( PSCR_IFD_TYPE )
                 );
	SmartCardExtension->VendorAttr.IfdType.Length = sizeof( PSCR_IFD_TYPE );
	


	SmartCardExtension->VendorAttr.UnitNo = pDevContext->DeviceInstanceNo;

    SmartCardExtension->VendorAttr.IfdVersion.BuildNumber = 0;

	 //   store firmware revision in ifd version
	SmartCardExtension->VendorAttr.IfdVersion.VersionMajor = 1;
	//ReaderExtension->FirmwareMajor;
	SmartCardExtension->VendorAttr.IfdVersion.VersionMinor = 1;
	//ReaderExtension->FirmwareMinor;
	SmartCardExtension->VendorAttr.IfdSerialNo.Length = 0;
	
	//   setup smartcard extension - reader capabilities
	KdPrint(("PcscRegisterWithSmcLib: Reader capabilites started \n"));
	SmartCardExtension->ReaderCapabilities.SupportedProtocols = SCARD_PROTOCOL_T0;
    SmartCardExtension->ReaderCapabilities.ReaderType = SCARD_READER_TYPE_SERIAL;
    SmartCardExtension->ReaderCapabilities.MechProperties = 0;
    SmartCardExtension->ReaderCapabilities.Channel = 0;
    SmartCardExtension->ReaderCapabilities.CLKFrequency.Default = 4000;
    SmartCardExtension->ReaderCapabilities.CLKFrequency.Max = 4000;
	SmartCardExtension->ReaderCapabilities.DataRate.Default = 10600;//	dataRatesSupported[0];
	SmartCardExtension->ReaderCapabilities.DataRate.Max = 10600;//	dataRatesSupported[2];
	SmartCardExtension->ReaderCapabilities.MaxIFSD = 254;
	SmartCardExtension->ReaderCapabilities.PowerMgmtSupport = 0;
	SmartCardExtension->ReaderCapabilities.CardConfiscated = FALSE;

	//   enter correct version of the lib
    SmartCardExtension->Version = SMCLIB_VERSION;
    SmartCardExtension->SmartcardRequest.BufferSize   = MIN_BUFFER_SIZE;
    SmartCardExtension->SmartcardReply.BufferSize  = MIN_BUFFER_SIZE;

    SmartCardExtension->ReaderExtension->ReaderPowerState = PowerReaderWorking;

    status = SmartcardInitialize(SmartCardExtension);//allocates internal buffers for the library

    if (status != STATUS_SUCCESS) {

        SmartcardLogError(
                         WdfDriverWdmGetDriverObject(WdfGetDriver()),
                         PSCR_INSUFFICIENT_RESOURCES,
                         NULL,
                         0
                         );
        return status;
    }

	SmartCardExtension->OsData->DeviceObject =	WdfDeviceWdmGetDeviceObject(pDevContext->Device);
	KdPrint(("PcscRegisterWithSmcLib: Assigned device object \n"));

    return status;
}





NTSTATUS
BthScrdEvtDeviceSelfManagedIoInit(
    __in WDFDEVICE  Device
    )
/*++
BthScrdEvtDeviceSelfManagedIoInit:
    This routine is called by the framework only once => used for one time initialization.
    In this routine local bth address, register server and publish SDP record  are retrieved.
--*/
{
    NTSTATUS status;
    PBTH_DEVICE_CONTEXT pDevContext = GetServerDeviceContext(Device);
    PUCHAR sdpRecordStream = NULL;
    ULONG sdpRecordLength = 0;
    BTHDDI_SDP_NODE_INTERFACE sdpNodeInterface;
    BTHDDI_SDP_PARSE_INTERFACE sdpParseInterface;

	PSMARTCARD_EXTENSION SmartcardExtension;
	UCHAR state;

	KdPrint(("=======>> BthScrdEvtDeviceSelfManagedIoInit: Enter \n"));

    status = BthEchoSharedRetrieveLocalBthAddress(pDevContext);
    if (!NT_SUCCESS(status)){
        return status;
    }

    status = BthEchoSrvRegisterPSM(pDevContext);
    if (!NT_SUCCESS(status)){
        return status;
    }

    status = BthEchoSrvRegisterL2CAPServer(pDevContext);
    if (!NT_SUCCESS(status)){
        return status;
    }

	//Profile drivers query for the GUID_BTHDDI_SDP_PARSE_INTERFACE to obtain
	//pointers to functions that allow them to parse SDP records.
    status = WdfFdoQueryForInterface(
        Device,
        &GUID_BTHDDI_SDP_PARSE_INTERFACE, 
        (PINTERFACE) (&sdpParseInterface),
        sizeof(sdpParseInterface), 
        BTHDDI_SDP_PARSE_INTERFACE_VERSION_FOR_QI, 
        NULL
        );
                
    if (!NT_SUCCESS(status))
    {
		KdPrint(("QueryInterface failed for Interface sdp parse interface, version %d, Status code %s\n", 
			BTHDDI_SDP_PARSE_INTERFACE_VERSION_FOR_QI, status));

        return status;
    }

	//Profile drivers query for the GUID_BTHDDI_SDP_NODE_INTERFACE to obtain
	//pointers to functions that allow them to create Service Discovery Protocol (SDP) records
    status = WdfFdoQueryForInterface(
        Device,
        &GUID_BTHDDI_SDP_NODE_INTERFACE, 
        (PINTERFACE) (&sdpNodeInterface),
        sizeof(sdpNodeInterface), 
        BTHDDI_SDP_NODE_INTERFACE_VERSION_FOR_QI, 
        NULL
        );
                
    if (!NT_SUCCESS(status))
    {
		KdPrint(("QueryInterface failed for Interface sdp node interface, version %d, Status code %s\n", 
			BTHDDI_SDP_NODE_INTERFACE_VERSION_FOR_QI, status));

        return status; 
    }

    status = CreateSdpRecord(
		&sdpNodeInterface,
        &sdpParseInterface,
        &BTHSCRD_SVC_GUID,
        BthScrdSvcName,
        pDevContext->Psm,
        &sdpRecordStream,
        &sdpRecordLength
        );                                                
    if (!NT_SUCCESS(status))
    {
	    if (sdpRecordStream)
			FreeSdpRecord(sdpRecordStream);
		return status;
    }
    
    status = BthEchoSrvPublishSdpRecord(pDevContext, sdpRecordStream, sdpRecordLength);
	KdPrint(("BthScrdEvtDeviceSelfManagedIoInit: BthEchoSrvPublishSdpRecordstatus, status =  %s\n", status));
    if (!NT_SUCCESS(status))
    {
	    if (sdpRecordStream)
			FreeSdpRecord(sdpRecordStream);
		return status;
    }

	KdPrint(("BthScrdEvtDeviceSelfManagedIoInit: SmartcardExtension part \n"));
	SmartcardExtension = (PSMARTCARD_EXTENSION) &pDevContext->SmartCardExtension;
    SmartcardExtension->ReaderExtension->StatusFileSelected = FALSE;
    state = PSCR_ICC_ABSENT;//CARD is absent
	CBUpdateCardState(SmartcardExtension, state, FALSE);

	KdPrint(("<<======= BthScrdEvtDeviceSelfManagedIoInit: Exit \n\n"));
    
    return status;    
}    





VOID
BthScrdEvtDeviceSelfManagedIoCleanup(
    __in WDFDEVICE  Device
    )
/*++
BthScrdEvtDeviceSelfManagedIoCleanup:
    This routine is called by the framework only once => used for one time de-initialization.
    Tasks: remove sdp record, unregister server, and disconnect all the outstanding connections.
--*/
{
	
    PBTH_DEVICE_CONTEXT pDevContext = GetServerDeviceContext(Device);
        
    PAGED_CODE();

	KdPrint(("BthScrdEvtDeviceSelfManagedIoCleanup: Enter \n"));

    if ( pDevContext->SdpRecordHandle != HANDLE_SDP_NULL)
        BthSrvRemoveSdpRecord(pDevContext);        

    if (pDevContext->L2CAPServerHandle != NULL)
        BthSrvUnregisterL2CAPServer(pDevContext);

    if (pDevContext->Psm != 0)
        BthSrvUnregisterPSM(pDevContext);

    /*
    Disconnect any open connections, after this point no more connections
	can come beacuse we have unregistered server. 
	BthEchoSrvDisconnectConnectionsOnRemove	does not wait for disconnect to
	complete. Connection object's cleanup callback waits on that. Since connection
	objects are children of device,	they will be cleaned up and disconnect
	would complete before device object is cleaned up.
	*/
    
    BthEchoSrvDisconnectConnectionsOnRemove(pDevContext);
	KdPrint(("BthScrdEvtDeviceSelfManagedIoCleanup: Exit \n"));
}




VOID
PscrEvtIoDeviceControl(
    IN WDFQUEUE    Queue,
    IN WDFREQUEST  Request,
    IN size_t      OutputBufferLength,
    IN size_t      InputBufferLength,
    IN ULONG       IoControlCode
    )
/*++
PscrEvtIoDeviceControl:
    This event is called when the framework receives IRP_MJ_DEVICE_CONTROL
    requests from the system.

Arguments:
    Queue - Handle to the framework queue object that is associated with the I/O request.
    Request - Handle to a framework request object.

    OutputBufferLength - length of the request's output buffer,
                        if an output buffer is available.
    InputBufferLength - length of the request's input buffer,
                        if an input buffer is available.

    IoControlCode - the driver-defined or system-defined I/O control code
                    (IOCTL) that is associated with the request.
Return Value:
--*/
{
    PBTH_DEVICE_CONTEXT pDevContext;
    PIRP irp;
	NTSTATUS status;

    UNREFERENCED_PARAMETER(OutputBufferLength);
    UNREFERENCED_PARAMETER(InputBufferLength);
    UNREFERENCED_PARAMETER(IoControlCode);

    KdPrint(("\n-----> PscrEvtIoDeviceControl called %s Request 0x%x\n",
                GetIoctlName(IoControlCode), Request));

    pDevContext = GetServerDeviceContext(WdfIoQueueGetDevice(Queue));

    //
    // Since smartcard library expects an IRP, we will get the underlying
    // irp.
    //
    irp = WdfRequestWdmGetIrp(Request);

    //
    // We will store the Request handle in the IRP DriverContext field so
    // that we can get the handle back in vendor-ioctl callback routine.
    //
    SET_REQUEST_IN_IRP(irp, Request);

    //
    // To workaround the problem of smartcard completing the IRP we provided.
    // we will use the extra stack location to set a completion routine so that
    // when it completes the IRP, our completion routine will be called, where
    // we can interrupt the completion and complete the actual WDFREQUEST.
    //
	KdPrint(("  SET_REQUEST_IN_IRP is called\n"));
	// copies the IRP stack parameters from the current I/O stack location
	// to the stack location of the next-lower driver
    IoCopyCurrentIrpStackLocationToNext(irp);

	//registers an IoCompletion routine, which will be called when the next-lower-level driver
	//has completed the requested operation for the given IRP
    IoSetCompletionRoutine (irp,
                           PscrSmcLibComplete,//our completition routine
                           Request,
                           TRUE,
                           TRUE,
                           TRUE
                           );

    IoSetNextIrpStackLocation(irp);//sets the IRP stack location in a driver-allocated IRP to our driver

    //
    // Ignore the return value because framework has marked the irp pending
    // and will return status-pending when we return from this routine.
    //
	/* SmartcardDeviceControl routine is the main entry routine for the
	   smart card driver library. It performs parameter checking and completes
	   calls that do not require I/O to or from the smart card reader */

    //(VOID) SmartcardDeviceControl(&(deviceExtension->SmartCardExtension), irp);
	
	//KdPrint((" PscrEvtIoDeviceControl: called complete_SmartCardIoctl \n"));
	status = complete_SmartCardIoctl( Queue, Request, OutputBufferLength, InputBufferLength, IoControlCode);

    return;
}




__drv_functionClass(IO_COMPLETION_ROUTINE)
NTSTATUS
PscrSmcLibComplete (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID Context
    )
/*++
Routine Description:
    Completion routine is called when the IRP is completed by the smclib.
    We will interrupt the completion of the IRP, and inturn complete the
    WDFREQUEST given to us by the framework.

--*/
{
    UNREFERENCED_PARAMETER(DeviceObject);

    KdPrint(("<---- PscrSmcLibComplete called %x Request %x\n", Irp->IoStatus.Status, Context));

    WdfRequestComplete((WDFREQUEST)Context, Irp->IoStatus.Status);

    return STATUS_MORE_PROCESSING_REQUIRED;
}







NTSTATUS
complete_SmartCardIoctl(
	 IN WDFQUEUE Queue,
	 IN WDFREQUEST Request,
	 IN size_t OutputBufferLength,
	 IN size_t InputBufferLength,
	 IN ULONG IoControlCode)
{
	NTSTATUS status;
	WDFDEVICE device;
	PBTH_DEVICE_CONTEXT pDevContext;
	PIRP Irp;

	device = WdfIoQueueGetDevice(Queue);
	pDevContext = GetServerDeviceContext(device);


	status = SmartcardAcquireRemoveLock(&pDevContext->SmartCardExtension);
	KdPrint(("  complete_SmartCardIoctl: Enter\n"));
	if(status != STATUS_SUCCESS)
	{
		KdPrint(("  complete_SmartCardIoctl status != STATUS_SUCCESS		\n"));
		WdfRequestCompleteWithInformation(Request, status, 0);
	}
	else
	{
		//KdPrint((" SmartcardAcquireRemoveLock STATUS_SUCCESS ==> \n Else entered\n"));
		Irp = WdfRequestWdmGetIrp(Request);
		KdPrint(("  The value of IRP = 0x%x \n",Irp));
		KdPrint(("  The value of SMART CARD EXTENSION = 0x%x \n",&pDevContext->SmartCardExtension));
		KdPrint(("  The value of IRP minor IOCTL code = 0x%x \n",Irp->AssociatedIrp.SystemBuffer));

		//SmartcardDeviceControl routine is the main entry routine for the smart card driver library.
		//It performs parameter checking and completes calls that do not require I/O to or from the smart card reader.
		//The driver's DeviceControl routine must call SmartcardDeviceControl routine so that the library
		//can check parameters and complete calls that do not require I/O operations to or from the smart card reader

		switch (IoControlCode)
		{
		case IOCTL_SMARTCARD_POWER:
			KdPrint(("  case IOCTL_SMARTCARD_POWER\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension, Irp);
			//WdfRequestComplete(Request, STATUS_SUCCESS);			
			break;
		case IOCTL_SMARTCARD_GET_ATTRIBUTE:
			KdPrint(("  case IOCTL_SMARTCARD_GET_ATTRIBUTE\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension,	Irp);
			break;
		case IOCTL_SMARTCARD_SET_ATTRIBUTE:
			KdPrint(("  case IOCTL_SMARTCARD_SET_ATTRIBUTE\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension,	Irp);
			break;
		case IOCTL_SMARTCARD_CONFISCATE:
			KdPrint(("  case IOCTL_SMARTCARD_CONFISCATE\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension,	Irp);
			break;
		case IOCTL_SMARTCARD_TRANSMIT:
			KdPrint(("  case IOCTL_SMARTCARD_TRANSMIT\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension, Irp);			
			break;
		case IOCTL_SMARTCARD_EJECT:
			KdPrint(("  case IOCTL_SMARTCARD_EJECT\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension,	Irp);
			break;
		case IOCTL_SMARTCARD_SWALLOW:
			KdPrint(("  case IOCTL_SMARTCARD_SWALLOW\n"));
			break;
		case IOCTL_SMARTCARD_IS_PRESENT:
			KdPrint(("  case IOCTL_SMARTCARD_IS_PRESENT\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension,	Irp);			
			break;
		case IOCTL_SMARTCARD_IS_ABSENT:
			KdPrint(("  case IOCTL_SMARTCARD_IS_ABSENT\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension,	Irp);
			break;
		case IOCTL_SMARTCARD_SET_PROTOCOL:
			KdPrint(("  case IOCTL_SMARTCARD_SET_PROTOCOL\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension,	Irp);
			break;
		case IOCTL_SMARTCARD_GET_STATE:
			KdPrint(("  case IOCTL_SMARTCARD_GET_STATE\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension,	Irp);
			break;
		case IOCTL_SMARTCARD_GET_LAST_ERROR:
			KdPrint(("  case IOCTL_SMARTCARD_GET_LAST_ERROR\n"));
			KdPrint(("  SmartcardDeviceControl is called\n"));
			status = SmartcardDeviceControl(&pDevContext->SmartCardExtension,	Irp);
			
			break;
		default:
			KdPrint(("  case default: UNKNOWN IOCTL\n"));
		}
		
		KdPrint(("  SmartcardReleaseRemoveLock is called	\n"));
		SmartcardReleaseRemoveLock(&pDevContext->SmartCardExtension);
		KdPrint(("  complete_SmartCardIoctl: Exit\n"));
	}
	
	return status;
}


PCHAR
GetIoctlName(
    ULONG IoControlCode
    )
{
    switch (IoControlCode) {
    case IOCTL_SMARTCARD_POWER: return "IOCTL_SMARTCARD_POWER";
    case IOCTL_SMARTCARD_GET_ATTRIBUTE: return "IOCTL_SMARTCARD_GET_ATTRIBUTE";
    case IOCTL_SMARTCARD_SET_ATTRIBUTE: return "IOCTL_SMARTCARD_SET_ATTRIBUTE";
    case IOCTL_SMARTCARD_CONFISCATE: return "IOCTL_SMARTCARD_CONFISCATE";
    case IOCTL_SMARTCARD_TRANSMIT: return "IOCTL_SMARTCARD_TRANSMIT";
    case IOCTL_SMARTCARD_EJECT: return "IOCTL_SMARTCARD_EJECT";
    case IOCTL_SMARTCARD_SWALLOW: return "IOCTL_SMARTCARD_SWALLOW";
    case IOCTL_SMARTCARD_IS_PRESENT: return "IOCTL_SMARTCARD_IS_PRESENT";
    case IOCTL_SMARTCARD_IS_ABSENT: return "IOCTL_SMARTCARD_IS_ABSENT";
    case IOCTL_SMARTCARD_SET_PROTOCOL: return "IOCTL_SMARTCARD_SET_PROTOCOL";
    case IOCTL_SMARTCARD_GET_STATE: return "IOCTL_SMARTCARD_GET_STATE";
    case IOCTL_SMARTCARD_GET_LAST_ERROR: return "IOCTL_SMARTCARD_GET_LAST_ERROR";
    default: return "Unknow_ioctl";
    }
}




VOID
PscrEvtIoCanceledOnQueue(
    IN WDFQUEUE  Queue,
    IN WDFREQUEST  Request
    )
/*++

PscrEvtIoCanceledOnQueue:
  This routine is called when the request is cancelled on the queue.

  We will use the notificatio to clear the NotificationIrp field set by
  the smartcard library.
--*/
{
	PBTH_DEVICE_CONTEXT pDevContext;
    PSMARTCARD_EXTENSION smartcardExtension;

    pDevContext = GetServerDeviceContext(WdfIoQueueGetDevice(Queue));
    smartcardExtension = (PSMARTCARD_EXTENSION) &pDevContext->SmartCardExtension;

    KdPrint(("PscrEvtIoCanceledOnQueue called for  0x%x\n", WdfRequestWdmGetIrp(Request)));

    InterlockedExchangePointer(
                             &(smartcardExtension->OsData->NotificationIrp),
                             NULL
                             );

    WdfRequestComplete(Request, STATUS_CANCELLED);
}




VOID
PcscEvtDeviceContextCleanup (
    WDFDEVICE       Device
    )
/*++
Routine Description:
    Called when the device object is deleted by the framework in response
    to REMOVE_DEVICE request. Here, free up all the resources that wouldn't
    cleaned up by the framework automatically due to parent child relationship.

Arguments:
    Device - Handle to a framework device object.
--*/
{
	PBTH_DEVICE_CONTEXT pDevContext;

	KdPrint(("\n----------------------------------------\n"));
	KdPrint(("PSCR! PscrEvtDeviceContextCleanup: Enter\n"));

    pDevContext = GetServerDeviceContext(Device);

    InterlockedDecrement((PLONG)&pDevContext->DeviceInstanceNo);

    if (pDevContext->SmartCardExtension.ReaderExtension != NULL) 
	{

        ExFreePoolWithTag(pDevContext->SmartCardExtension.ReaderExtension,
                            SMARTCARD_POOL_TAG);

        pDevContext->SmartCardExtension.ReaderExtension = NULL;
    }
}