//--------------------------------------------------

#include "connection.h"



__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoRepeatReaderSubmit(
    __in PBTHECHO_CONNECTION Connection,
    __in PBTHECHO_REPEAT_READER RepeatReader
    );

__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthEchoRepeatReaderWaitForStop(    
    __in PBTHECHO_REPEAT_READER RepeatReader
    );

__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthEchoConnectionObjectWaitForAndUninitializeContinuousReader(
    __in PBTHECHO_CONNECTION Connection
    );


#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, BthEchoRepeatReaderWaitForStop)
#pragma alloc_text (PAGE, BthEchoConnectionObjectWaitForAndUninitializeContinuousReader)
#pragma alloc_text (PAGE, BthEchoEvtConnectionObjectCleanup)
#pragma alloc_text (PAGE, BthEchoConnectionObjectRemoteDisconnectSynchronously)
#endif



KDEFERRED_ROUTINE
BthEchoRepeatReaderResubmitReadDpc;

VOID
BthEchoRepeatReaderResubmitReadDpc(
    __in struct _KDPC  *Dpc,
    __in_opt PVOID  DeferredContext,
    __in_opt PVOID  SystemArgument1,
    __in_opt PVOID  SystemArgument2
)
/*++
Description: Dpc for resubmitting pending read

Arguments:
    Dpc - Dpc
    DeferredContext - We receive device context header as DeferredContext
    SystemArgument1 - We receive repeat reader as SystemArgument1
    SystemArgument2 - Unused
--*/
{
    PBTHECHO_REPEAT_READER repeatReader;
    
    UNREFERENCED_PARAMETER(Dpc);
    UNREFERENCED_PARAMETER(SystemArgument2);

	KdPrint(("SCRD! BthEchoRepeatReaderResubmitReadDpc: Enter\n"));

    repeatReader = (PBTHECHO_REPEAT_READER) SystemArgument1;

    //
    // BthEchoRepeatReaderSubmit will invoke contreader failure callback
	//in case of failure so we don't check for failure here
    //
    (void) BthEchoRepeatReaderSubmit(
		(PBTHECHO_CONNECTION)DeferredContext,
        repeatReader
        );
}





EVT_WDF_REQUEST_COMPLETION_ROUTINE
BthEchoRepeatReaderPendingReadCompletion;

void
BthEchoRepeatReaderPendingReadCompletion(
    __in WDFREQUEST  Request,
    __in WDFIOTARGET  Target,
    __in PWDF_REQUEST_COMPLETION_PARAMS  Params,
    __in WDFCONTEXT  Context
    )
/*++
Description:
    Completion routine for pending read request

    In this routine we invoke the read completion callback in case of success
	and contreader failure callback in case of failure.
    These are implemented by server.

Arguments:
    Request - Request completed
    Target - Target to which request was sent
    Params - Completion parameters
    Context - We receive repeat reader as the context
--*/
{
    PBTHECHO_REPEAT_READER repeatReader;
    NTSTATUS status;

    UNREFERENCED_PARAMETER(Target);
    UNREFERENCED_PARAMETER(Request);

	KdPrint(("SCRD! BthEchoRepeatReaderPendingReadCompletion: Enter\n"));


    repeatReader = (PBTHECHO_REPEAT_READER) Context;
       
    ASSERT((repeatReader != NULL));

    status = Params->IoStatus.Status;

    if (NT_SUCCESS(status))
    {
		KdPrint( ("SCRD! Pending read completion, RepeatReader: 0x%p, status: %s, Buffer: 0x%p, BufferSize: %d\n",
			repeatReader,
            Params->IoStatus.Status,
            repeatReader->TransferBrb.Buffer,
            repeatReader->TransferBrb.BufferSize            
            ));
           
        repeatReader->Connection->ContinuousReader.BthEchoConnectionObjectContReaderReadCompleteCallback(
            repeatReader->Connection,
            repeatReader->TransferBrb.Buffer,
            repeatReader->TransferBrb.BufferSize
            );

    }
    else
    {
		KdPrint (( "!!!!! Pending read completed with failure, RepeatReader: 0x%p, status: %x\n", 
            repeatReader,
            Params->IoStatus.Status          
            ));
    }

    if (!NT_SUCCESS(status))
    {
        //
        // Invoke reader failed callback only if status is something other
        // than cancelled. If the status is STATUS_CANCELLED this is because
        // we have cancelled repeat reader.
        //
        if (STATUS_CANCELLED != status)
        {       
            //
            // Invoke reader failed callback before setting the stop event
            // to ensure that connection object is alive during this callback
            //
            repeatReader->Connection->ContinuousReader.BthEchoConnectionObjectContReaderFailedCallback(
                repeatReader->Connection
                );
        }
            
        //
        // Set stop event because we are not resubmitting the reader
        //
        KeSetEvent(&repeatReader->StopEvent, 0, FALSE);        
    }
    else
    {
        //
        // Resubmit pending read
        // We use a Dpc to avoid recursing in this completion routine
        //

        BOOLEAN ret = KeInsertQueueDpc(&repeatReader->ResubmitDpc, repeatReader, NULL);
        ASSERT (TRUE == ret); //we only expect one outstanding dpc
        UNREFERENCED_PARAMETER(ret); //ret remains unused in fre build
    }
}




VOID
BthEchoRepeatReaderUninitialize(    
    __in PBTHECHO_REPEAT_READER RepeatReader
    )
{
	KdPrint(("SCRD! BthEchoRepeatReaderUninitialize: Enter\n"));
    if (NULL != RepeatReader->RequestPendingRead)
    {
        RepeatReader->RequestPendingRead = NULL;
        RepeatReader->Stopping = 0;        
    }    
}




__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoRepeatReaderInitialize(
    __in PBTHECHO_CONNECTION Connection,
    __in PBTHECHO_REPEAT_READER RepeatReader,
    __in size_t BufferSize    
    )

/*++
Description:    This routine initializes repeat reader.

Arguments:
    Connection - Connection with which this repeat reader is associated
    RepeatReader - Repeat reader
    BufferSize - Buffer size for read
--*/

{
    NTSTATUS status;
    WDF_OBJECT_ATTRIBUTES attributes;

	KdPrint(("SCRD! BthEchoRepeatReaderInitialize: Enter\n"));

    //
    // Create request object for pending read
    // Set connection object as the parent for the request
    //
    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.ParentObject = WdfObjectContextGetObject(Connection);

    status = WdfRequestCreate(
        &attributes,
        Connection->IoTarget,
        &RepeatReader->RequestPendingRead
        );                    

    if (!NT_SUCCESS(status))
    {
		KdPrint (("Creating request for pending read failed, "
            "Status code %x\n",
            status
            ));

        goto exit;
    }

    //
    // Create memory object for the pending read
    //
    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.ParentObject = RepeatReader->RequestPendingRead;

    status = WdfMemoryCreate(
        &attributes,
        NonPagedPool,
        POOLTAG_BTHECHOSAMPLE,
        BufferSize,
        &RepeatReader->MemoryPendingRead,
        NULL //buffer
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint (("Creating memory for pending read failed, Status code %s\n", status ));

        goto exit;
    }

    //
    // Initialize Dpc that we will use for resubmitting pending read
    // Deferred procedure calls (DPCs), which can be queued from ISRs and which
	// are executed at a later time and at a lower IRQL than the ISR
    KeInitializeDpc(
        &RepeatReader->ResubmitDpc, 
        BthEchoRepeatReaderResubmitReadDpc, 
        Connection // the value to pass as the DeferredContext parameter to BthEchoRepeatReaderResubmitReadDpc
        );

    //
    // Initialize event used to wait for stop.
    // This even is created as signalled. It gets cleared when
    // request is submitted.
    //
    KeInitializeEvent(&RepeatReader->StopEvent, NotificationEvent, TRUE);

    RepeatReader->Connection = Connection;
    
exit:
    if (!NT_SUCCESS(status))
    {
        BthEchoRepeatReaderUninitialize(RepeatReader);
    }
    
    return status;
}




__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoRepeatReaderSubmit(
    __in PBTHECHO_CONNECTION Connectio,
    __in PBTHECHO_REPEAT_READER RepeatReader
    )
/*++
Description:
    This routine submits the repeat reader.
    In case of failure it invoked contreader failed callback

Arguments:
    Connectio ~ Device context header
    RepeatReader - Repeat reader to submit
--*/
{
    NTSTATUS status, statusReuse;
    WDF_REQUEST_REUSE_PARAMS reuseParams;
    struct _BRB_L2CA_ACL_TRANSFER *brb = &RepeatReader->TransferBrb; 

	KdPrint(("SCRD! BthEchoRepeatReaderSubmit: Enter\n"));

	//reinitializes a Bluetooth request block (BRB) to be reused
    Connectio->ProfileDrvInterface.BthReuseBrb((PBRB)brb, BRB_L2CA_ACL_TRANSFER);

    WDF_REQUEST_REUSE_PARAMS_INIT(&reuseParams, WDF_REQUEST_REUSE_NO_FLAGS, STATUS_UNSUCCESSFUL);
    statusReuse = WdfRequestReuse(RepeatReader->RequestPendingRead, &reuseParams);
    ASSERT(NT_SUCCESS(statusReuse));
    UNREFERENCED_PARAMETER(statusReuse);

    //
    // Check if we are stopping, if yes set StopEvent and exit.
    //
    // After this point request is eligible for cancellation, so if this
    // flag gets set after we check it, request will be cancelled for stopping
    // the repeat reader and next time around we will stop when we are invoked
    // again upon completion of cancelled request.
    //
    if (RepeatReader->Stopping)
    {
		KdPrint(("SCRD! BthEchoRepeatReaderSubmit: Continuos reader 0x%p stopping\n", RepeatReader));
        
        KeSetEvent(&RepeatReader->StopEvent, 0, FALSE);

        status = STATUS_SUCCESS;

        goto exit;
    }

    //
    // Format request for L2CA IN transfer
    //
    status = BthEchoConnectionObjectFormatRequestForL2CaTransfer(
        RepeatReader->Connection,
        RepeatReader->RequestPendingRead,
        &brb,
        RepeatReader->MemoryPendingRead,
        ACL_TRANSFER_DIRECTION_IN | ACL_SHORT_TRANSFER_OK
        );

    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    //
    // Set a CompletionRoutine callback function. 
	// framework passes the RepeatReader to the CompletionRoutine callback function
    //
    WdfRequestSetCompletionRoutine(
        RepeatReader->RequestPendingRead, //A handle to a framework request object
        BthEchoRepeatReaderPendingReadCompletion,
        RepeatReader
        );

    //
    // Clear the stop event before sending the request. This is relevant only
	// on start of the repeat reader (i.e. the first submission)
    // this event eventually gets set only when repeat reader stops
    // and not on every resubmission.
    //
    KeClearEvent(&RepeatReader->StopEvent);

    if (FALSE == WdfRequestSend(
        RepeatReader->RequestPendingRead,
        Connectio->IoTarget,
        NULL
        ))
    {
        status = WdfRequestGetStatus(RepeatReader->RequestPendingRead);

		KdPrint (("!!!!! Request send failed for request 0x%p, Brb 0x%p, Status code %s\n", 
            RepeatReader->RequestPendingRead,
            brb,
            status
            ));

        goto exit;
    }
    else
    {
		KdPrint(("SCRD! BthEchoRepeatReaderSubmit: Resubmited pending read with request 0x%p, Brb 0x%p\n", 
            RepeatReader->RequestPendingRead,
            brb
            ));
    }

exit:
    if (!NT_SUCCESS(status))
    {
        //
        // Invoke the reader failed callback before setting the event
        // to ensure that the connection object is alive during this callback
        //
        RepeatReader->Connection->ContinuousReader.BthEchoConnectionObjectContReaderFailedCallback(
            RepeatReader->Connection
            );

        //
        // If we failed to send pending read, set the event since
        // we will not get completion callback
        //
            
        KeSetEvent(&RepeatReader->StopEvent, 0, FALSE);
    }

	KdPrint(("SCRD! BthEchoRepeatReaderSubmit: Exit\n"));
    
    return status;
}



__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoRepeatReaderCancel(
    __in PBTHECHO_REPEAT_READER RepeatReader    
    )
/*++
Description:
    This routine cancels repeat reader but it does not wait for repeat reader to stop.
    This wait must be performed separately.

Arguments:
    RepeatReader - Repeat reader to be cancelled
--*/
{
    //
    // Signal that we are transitioning to stopped state so that we don't resubmit pending read
    //
	KdPrint(("SCRD! BthEchoRepeatReaderCancel: Enter\n"));
    
    InterlockedIncrement(&RepeatReader->Stopping);

    //
    // Cancel the pending read. If BthEchoRepeatReaderSubmit misses the setting of Stopping
    // flag, cancel would cause the request to complete quickly and repeat reader will
	// subsequently be stopped.
    //
    WdfRequestCancelSentRequest(RepeatReader->RequestPendingRead);    
}


__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthEchoRepeatReaderWaitForStop(    
    __in PBTHECHO_REPEAT_READER RepeatReader
    )
/*++
Description:    This routine waits for repeat reader to stop
--*/
{
    PAGED_CODE();

    //
    // Wait for pending read to complete
    //

	KdPrint(("SCRD! BthEchoRepeatReaderWaitForStop: Enter\n"));
    
    KeWaitForSingleObject(
        &RepeatReader->StopEvent,
        Executive,
        KernelMode,
        FALSE,
        NULL
        );
	KdPrint(("SCRD! BthEchoRepeatReaderWaitForStop: Exit\n"));
}




//====================================================================
//Connection object functions
//====================================================================


__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoConnectionObjectInitializeContinuousReader(
    __in PBTHECHO_CONNECTION Connection,
    __in PFN_BTHECHO_CONNECTION_OBJECT_CONTREADER_READ_COMPLETE 
        BthEchoConnectionObjectContReaderReadCompleteCallback,
    __in PFN_BTHECHO_CONNECTION_OBJECT_CONTREADER_FAILED 
        BthEchoConnectionObjectContReaderFailedCallback,
    __in size_t BufferSize
    )
/*++
Description: This routine initializes continuous reader for connection

Arguments:
    Connection - Connection for which to initialize continuous readers
    BthEchoConnectionObjectContReaderReadCompleteCallback - Pending read completion callback
    BthEchoConnectionObjectContReaderFailedCallback - Repeat reader failed callback
    BufferSize - Buffer size for the pending read
--*/

{
    NTSTATUS status = STATUS_SUCCESS;
    UINT i;

	KdPrint(("SCRD! BthEchoConnectionObjectInitializeContinuousReader: Enter\n"));

    Connection->ContinuousReader.BthEchoConnectionObjectContReaderReadCompleteCallback = 
        BthEchoConnectionObjectContReaderReadCompleteCallback;
    Connection->ContinuousReader.BthEchoConnectionObjectContReaderFailedCallback = 
        BthEchoConnectionObjectContReaderFailedCallback;
        
    for (i = 0; i < BTHECHOSAMPLE_NUM_CONTINUOUS_READERS; i++)
    {
        status = BthEchoRepeatReaderInitialize(
            Connection,
            &Connection->ContinuousReader.RepeatReaders[i],
            BufferSize
            );
        
        if (!NT_SUCCESS(status))
        {
            break;
        }
        else
        {
            ++Connection->ContinuousReader.InitializedReadersCount;
        }
    }

    //
    // In case of error (as well as on success) BthEchoEvtConnectionObjectCleanup 
    // will uninitialize readers
    //
	KdPrint(("SCRD! BthEchoConnectionObjectInitializeContinuousReader: Exit\n"));
    
    return status;
}



__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoConnectionObjectContinuousReaderSubmitReaders(
    __in PBTHECHO_CONNECTION Connection
    )
/*++
Description:
    This routine submits all the initialized repeat readers for the connection

Arguments:
    Connection - Connection whose repeat readers are to be submitted
--*/
{
    NTSTATUS status = STATUS_SUCCESS;
    UINT i;

	KdPrint(("SCRD! BthEchoConnectionObjectContinuousReaderSubmitReaders: Enter\n"));

    ASSERT (Connection->ContinuousReader.InitializedReadersCount
        <= BTHECHOSAMPLE_NUM_CONTINUOUS_READERS);
    
    for (i = 0; i < Connection->ContinuousReader.InitializedReadersCount; i++)
    {
        status = BthEchoRepeatReaderSubmit(
			Connection,
            &Connection->ContinuousReader.RepeatReaders[i]
            );
        
        if (!NT_SUCCESS(status))
        {
			// Cancel any submitted continuous readers
			// We make sure that it is safe to call cancel on unsubmitted readers
	        
			BthEchoConnectionObjectContinuousReaderCancelReaders(Connection);
			break;
        }
    }
	KdPrint(("SCRD! BthEchoConnectionObjectContinuousReaderSubmitReaders: Exit\n"));
    
    return status;
}


__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoConnectionObjectContinuousReaderCancelReaders(
    __in PBTHECHO_CONNECTION Connection
    )
/*++
Description:
    This routine cancels all the initialized repeat readers for the connection

Arguments:
    Connection - Connection whose repeat readers are to be camcelled
--*/
{
    UINT i;

	KdPrint(("SCRD! BthEchoConnectionObjectContinuousReaderCancelReaders: Enter\n"));

    ASSERT (Connection->ContinuousReader.InitializedReadersCount
        <= BTHECHOSAMPLE_NUM_CONTINUOUS_READERS);

    for (i = 0; i < Connection->ContinuousReader.InitializedReadersCount; i++)
    {
        BthEchoRepeatReaderCancel(&Connection->ContinuousReader.RepeatReaders[i]);
    }
}



__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthEchoConnectionObjectWaitForAndUninitializeContinuousReader(
    __in PBTHECHO_CONNECTION Connection
    )
/*++
Description:
    This routine waits for and uninitializes all the initialized 
    continuous readers for the connection.

Arguments:
    Connection - Connection whose continuous reader is to be uninitialized
--*/
{
    UINT i;

    PAGED_CODE();

	KdPrint(("SCRD! BthEchoConnectionObjectWaitForAndUninitializeContinuousReader: Enter\n"));

    ASSERT (Connection->ContinuousReader.InitializedReadersCount
        <= BTHECHOSAMPLE_NUM_CONTINUOUS_READERS);
    
    for (i = 0; i < Connection->ContinuousReader.InitializedReadersCount; i++)
    {
        PBTHECHO_REPEAT_READER repeatReader = &Connection->ContinuousReader.RepeatReaders[i];

        BthEchoRepeatReaderWaitForStop(repeatReader); //It is OK to wait on unsubmitted readers
        BthEchoRepeatReaderUninitialize(repeatReader);
    }    
	KdPrint(("SCRD! BthEchoConnectionObjectWaitForAndUninitializeContinuousReader: Exit\n"));
}



VOID
BthEchoEvtConnectionObjectCleanup(
    __in WDFOBJECT  ConnectionObject
    )
/*++
Description:
    This routine is invoked by the Framework when connection object gets deleted
	(either explicitly or implicitly because of parent deletion).

    Since we mark ExecutionLevel as passive for connection object we get this
    callback at passive level and can wait for stopping of continuous
    readers and for disconnect to complete.
--*/
{
    PBTHECHO_CONNECTION connection = GetConnectionObjectContext(ConnectionObject);

    PAGED_CODE();

	KdPrint(("SCRD! BthEchoEvtConnectionObjectCleanup: Enter\n"));
        
    BthEchoConnectionObjectWaitForAndUninitializeContinuousReader(connection);

    KeWaitForSingleObject(&connection->DisconnectEvent,
        Executive,
        KernelMode,
        FALSE,
        NULL);

    WdfObjectDelete(connection->ConnectDisconnectRequest);

	KdPrint(("SCRD! BthEchoEvtConnectionObjectCleanup: Exit\n"));
}




__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoConnectionObjectInit(
    __in WDFOBJECT ConnectionObject,
    __in PBTH_DEVICE_CONTEXT pDevContext
    )
/*++
Description:
    This routine initializes connection object. It is invoked by BthEchoConnectionObjectCreate.

Arguments:
    ConnectionObject - Object to initialize
    pDevContext - Device context
--*/
{
    NTSTATUS status;
    WDF_OBJECT_ATTRIBUTES attributes;
    PBTHECHO_CONNECTION connection = GetConnectionObjectContext(ConnectionObject);

	KdPrint(("SCRD!    BthEchoConnectionObjectInit: Enter\n"));
    
    // ---connection->DevCtx = DevCtxHd;
	connection->Device = pDevContext->Device;
	connection->IoTarget = pDevContext->IoTarget;
	connection->ProfileDrvInterface = pDevContext->ProfileDrvInterface;
	connection->LocalBthAddr = pDevContext->LocalBthAddr;
	connection->Request = pDevContext->Request;
	// ---end

    connection->ConnectionState = ConnectionStateInitialized;

    //
    // Initialize spinlock
    //    
    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.ParentObject = ConnectionObject;

    status = WdfSpinLockCreate(
		&attributes,
		&connection->ConnectionLock
		);

    if (!NT_SUCCESS(status))
        return status;                

    //
    // Create connect/disconnect request
	//
    status = WdfRequestCreate(
        &attributes,
        pDevContext->IoTarget,
        &connection->ConnectDisconnectRequest
        );

    if (!NT_SUCCESS(status))
        return status;

    //
    // Initialize event
	//
    KeInitializeEvent(&connection->DisconnectEvent, NotificationEvent, TRUE);

    //
    // Initalize list entry
    //
    InitializeListHead(&connection->ConnectionListEntry);

    connection->ConnectionState = ConnectionStateInitialized;

	KdPrint(("SCRD!    BthEchoConnectionObjectInit: Exit\n"));

    return status;
}




__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoConnectionObjectCreate(
    __in PBTH_DEVICE_CONTEXT pDevContext,
    __in WDFOBJECT ParentObject,
    __out WDFOBJECT*  ConnectionObject
    )
/*++
Description:
    This routine creates a connection object.
    This is called by the server when a remote connection is made.

Arguments:
    DevCtxHdr - Device context header
    ParentObject - Parent object for connection object
    ConnectionObject - receives the created connection object
--*/
{
    NTSTATUS status;
    WDF_OBJECT_ATTRIBUTES attributes;
    WDFOBJECT connectionObject = NULL;

	KdPrint(("SCRD!    BthEchoConnectionObjectCreate: Enter\n"));

    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes, BTHECHO_CONNECTION);
    attributes.ParentObject = ParentObject;    
    attributes.EvtCleanupCallback = BthEchoEvtConnectionObjectCleanup;

    //
    // We set execution level to passive so that we get cleanup at passive
    // level where we can wait for continuous readers to run down
    // and for completion of disconnect
    //
    attributes.ExecutionLevel = WdfExecutionLevelPassive; 

    status = WdfObjectCreate(
        &attributes,
        &connectionObject
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint(("WdfObjectCreate for connection object failed, Status code %x\n", status));
        
        return status;
    }

    status = BthEchoConnectionObjectInit(connectionObject, pDevContext);

    if (!NT_SUCCESS(status))
    {
		KdPrint(("Context initialize for connection object failed, ConnectionObject 0x%p, Status code %s\n",
            connectionObject,
            status
            ));

        if(connectionObject)
			WdfObjectDelete(connectionObject);
    }

    *ConnectionObject = connectionObject;

	KdPrint(("SCRD!    BthEchoConnectionObjectCreate: Exit\n"));
    
    return status;
}



EVT_WDF_REQUEST_COMPLETION_ROUTINE
BthEchoConnectionObjectDisconnectCompletion;

void
BthEchoConnectionObjectDisconnectCompletion(
    __in WDFREQUEST  Request,
    __in WDFIOTARGET  Target,
    __in PWDF_REQUEST_COMPLETION_PARAMS  Params,
    __in WDFCONTEXT  Context
    )
/*++
Description:
    Completion routine for disconnect BRB completion. In this routine we set
	the connection to disconnect and set the DisconnectEvent.

Arguments:
    Request - Request completed
    Target - Target to which request was sent
    Params - Completion parameters
    Context - We receive connection as the context
--*/
{
    PBTHECHO_CONNECTION connection = (PBTHECHO_CONNECTION) Context;

    UNREFERENCED_PARAMETER(Request);
    UNREFERENCED_PARAMETER(Target);
    UNREFERENCED_PARAMETER(Params);

	KdPrint(("SCRD! BthEchoConnectionObjectDisconnectCompletion: Enter\n"));

    WdfSpinLockAcquire(connection->ConnectionLock);
    connection->ConnectionState = ConnectionStateDisconnected;
    WdfSpinLockRelease(connection->ConnectionLock);

    //
    // Disconnect complete, set the event
    //

    KeSetEvent(
        &connection->DisconnectEvent,
        0,
        FALSE
        );    
}




__drv_maxIRQL(DISPATCH_LEVEL)
BOOLEAN
BthEchoConnectionObjectRemoteDisconnect(
    __in WDFIOTARGET IoTarget,
	__in BTH_PROFILE_DRIVER_INTERFACE * ProfileDrvInterface,
    __in PBTHECHO_CONNECTION Connection
    )
/*++
Description:
    This routine sends a disconnect BRB for the connection

Arguments:
    DevCtxHdr - Device context header
    Connection - Connection which is to be disconnected

Return Value:
    TRUE is this call initiates the disconnect.
    FALSE if the connection was already disconnected.
--*/
{
    struct _BRB_L2CA_CLOSE_CHANNEL * disconnectBrb;

	KdPrint(("SCRD! BthEchoConnectionObjectRemoteDisconnect: Enter\n"));

    //
    // Cancel continuous readers for the connection
    //
    BthEchoConnectionObjectContinuousReaderCancelReaders(Connection);
    
    WdfSpinLockAcquire(Connection->ConnectionLock);
    
    if (Connection->ConnectionState == ConnectionStateConnecting) 
    {
        //
        // If the connection is not completed yet set the state 
        // to disconnecting.
        // In such case we should send CLOSE_CHANNEL Brb down after 
        // we receive connect completion.
        //
        
        Connection->ConnectionState = ConnectionStateDisconnecting;

        //
        // Clear event to indicate that we are in disconnecting
        // state. It will be set when disconnect is completed
        //
        KeClearEvent(&Connection->DisconnectEvent);

        WdfSpinLockRelease(Connection->ConnectionLock);
        return TRUE;

    } 
    else if (Connection->ConnectionState != ConnectionStateConnected)
    {
        //
        // Do nothing if we are not connected
        //

        WdfSpinLockRelease(Connection->ConnectionLock);
        return FALSE;
    }

    Connection->ConnectionState = ConnectionStateDisconnecting;
    WdfSpinLockRelease(Connection->ConnectionLock);

    //
    // We are now sending the disconnect, so clear the event.
    //

    KeClearEvent(&Connection->DisconnectEvent);

    ProfileDrvInterface->BthReuseBrb(&Connection->ConnectDisconnectBrb, BRB_L2CA_CLOSE_CHANNEL);

    disconnectBrb = (struct _BRB_L2CA_CLOSE_CHANNEL *) &(Connection->ConnectDisconnectBrb);
    
    disconnectBrb->BtAddress = Connection->RemoteAddress;
    disconnectBrb->ChannelHandle = Connection->ChannelHandle;

    //
    // The BRB can fail with STATUS_DEVICE_DISCONNECT if the device is already
    // disconnected, hence we don't assert for success
    //

    (void) BthEchoSharedSendBrbAsync(
        IoTarget, 
        Connection->ConnectDisconnectRequest, 
        (PBRB) disconnectBrb,
        sizeof(*disconnectBrb),
        BthEchoConnectionObjectDisconnectCompletion,
        Connection
        );    

    return TRUE;
}



__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthEchoConnectionObjectRemoteDisconnectSynchronously(
    __in WDFIOTARGET IoTarget,
	__in BTH_PROFILE_DRIVER_INTERFACE * ProfileDrvInterface,
    __in PBTHECHO_CONNECTION Connection
    )
/*++

Description:
    This routine disconnects the connection synchronously

Arguments:
    DevCtxHdr - Device context header
    Connection - Connection which is to be disconnected
--*/
{
    PAGED_CODE();

	KdPrint(("SCRD! BthEchoConnectionObjectRemoteDisconnectSynchronously: Enter\n"));

    BthEchoConnectionObjectRemoteDisconnect(IoTarget, ProfileDrvInterface, Connection);

    KeWaitForSingleObject(&Connection->DisconnectEvent,
        Executive,
        KernelMode,
        FALSE,
        NULL
        );    
}




__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
FormatRequestWithBrb(
    __in WDFIOTARGET IoTarget,
    __in WDFREQUEST Request,
    __in PBRB Brb,
    __in size_t BrbSize
    )
/*++
Description:
    This routine formats are WDFREQUEST with the passed in BRB

Arguments:
    IoTarget - Target to which request will be sent
    Request - Request to be formattted
    Brb - BRB to format the request with
    BrbSize - size of the BRB
--*/
{
    NTSTATUS status;
    WDF_OBJECT_ATTRIBUTES attributes;
    WDFMEMORY memoryArg1;

	KdPrint(("SCRD! FormatRequestWithBrb: Enter\n"));

    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.ParentObject = Request;

    status = WdfMemoryCreatePreallocated(
        &attributes,
        Brb,
        BrbSize,
        &memoryArg1
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint(("Creating preallocted memory for Brb 0x%p failed, Request to be formatted 0x%p, "
            "Status code %s\n",
            Brb,
            Request,
            status
            ));

        return status;
    }

    status = WdfIoTargetFormatRequestForInternalIoctlOthers(
        IoTarget,
        Request,
        IOCTL_INTERNAL_BTH_SUBMIT_BRB,
        memoryArg1,
        NULL, //OtherArg1Offset
        NULL, //OtherArg2
        NULL, //OtherArg2Offset
        NULL, //OtherArg4
        NULL  //OtherArg4Offset
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint(("Formatting request 0x%p with Brb 0x%p failed, Status code %s\n",
            Request,
            Brb,
            status
            ));
    }
    return status;
}



__drv_maxIRQL(DISPATCH_LEVEL)
__drv_when(return==0, __drv_allocatesMem(Mem))
NTSTATUS
BthEchoConnectionObjectFormatRequestForL2CaTransfer(
    __in PBTHECHO_CONNECTION Connection,
    __in WDFREQUEST Request,
    __inout struct _BRB_L2CA_ACL_TRANSFER ** Brb,
    __in WDFMEMORY Memory,
    __in ULONG TransferFlags //flags include direction of transfer
    )
/*++
Description: Formats a request for L2Ca transfer

Arguments:
    Connection - Connection on which L2Ca transfer will be made
    Request - Request to be formatted
    Brb - If a Brb is passed in, it will be used, otherwise
          this routine will allocate the Brb and return in this parameter
    Memory - Memory object which has the buffer for transfer
    TransferFlags - Transfer flags which include direction of the transfer
--*/
{
    NTSTATUS status = STATUS_SUCCESS;
    struct _BRB_L2CA_ACL_TRANSFER *brb = NULL;
    size_t bufferSize;
    BOOLEAN brbAllocatedLocally = FALSE; //whether this function allocated the BRB

	KdPrint(("SCRD! BthEchoConnectionObjectFormatRequestForL2CaTransfer: Enter\n"));

    WdfSpinLockAcquire(Connection->ConnectionLock);

    if(Connection->ConnectionState != ConnectionStateConnected)
    {
        status = STATUS_CONNECTION_DISCONNECTED;
        WdfSpinLockRelease(Connection->ConnectionLock);
        goto exit;
    }

    WdfSpinLockRelease(Connection->ConnectionLock);

    if (NULL == *Brb)
    {
        brb= (struct _BRB_L2CA_ACL_TRANSFER *)
            Connection->ProfileDrvInterface.BthAllocateBrb(
                BRB_L2CA_ACL_TRANSFER, 
                POOLTAG_BTHECHOSAMPLE
                );
        if(!brb)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;

			KdPrint(("Failed to allocate BRB_L2CA_ACL_TRANSFER, returning status code %x\n", status));

            goto exit;
        }
        else
        {
            brbAllocatedLocally = TRUE;
        }
    }
    else
    {
        brb = *Brb;
        Connection->ProfileDrvInterface.BthReuseBrb(
            (PBRB)brb, BRB_L2CA_ACL_TRANSFER
            );
    }

    brb->BtAddress = Connection->RemoteAddress;
    brb->BufferMDL = NULL;
    brb->Buffer = WdfMemoryGetBuffer(Memory, &bufferSize);

    if (bufferSize > (ULONG)(-1))
    {
        status = STATUS_BUFFER_OVERFLOW;

		KdPrint(("Buffer passed in longer than max ULONG, returning status code %x\n", status));

        goto exit;        
    }
    
    brb->BufferSize = (ULONG) bufferSize;
    brb->ChannelHandle = Connection->ChannelHandle;
    brb->TransferFlags = TransferFlags;

    status = FormatRequestWithBrb(
        Connection->IoTarget,
        Request,
        (PBRB) brb,
        sizeof(*brb)
        );
    
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    if (NULL == *Brb)
    {
        *Brb = brb;
    }
exit:
    if (!NT_SUCCESS(status) && brb && brbAllocatedLocally)
    {
        Connection->ProfileDrvInterface.BthFreeBrb((PBRB)brb);
    }
	KdPrint(("SCRD! BthEchoConnectionObjectFormatRequestForL2CaTransfer: Exit\n"));
    
    return status;
}









