/*++

Copyright (c) Microsoft Corporation.  All rights reserved.

    THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
    KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
    PURPOSE.

Module Name:
    Sdp.c

Abstract:
    Contains functionality for publishing and freeing SDP record
    
Environment:
    Kernel mode only
--*/


#include "sdp.h"
#include "connection.h"

#include <ntddk.h>
#include <wdf.h>
#include <initguid.h> 
#include <ntstrsafe.h>
#include <bthdef.h>
#include <ntintsafe.h>
#include <bthguid.h>
#include <bthioctl.h>
#include <sdpnode.h>
#include <bthddi.h>
#include <bthsdpddi.h>
#include <bthsdpdef.h>

#define POOLTAG_BTHECHOSAMPLE 'htbw'


NTSTATUS
AddSeqAttribute(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_TREE_ROOT_NODE Tree,
    __in USHORT AttribId,
    __out PSDP_NODE * Seq
    );

NTSTATUS
AppendSeqNode(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_NODE containerNode,
    __out PSDP_NODE * Node
    );

NTSTATUS
AppendNodeUint16(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_NODE ContainerNode,
    __in UINT16 Value
    );

NTSTATUS
AppendNodeUuid128(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_NODE ContainerNode,
    __in const GUID * Value
    );

NTSTATUS
AppendNodeUuid16(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_NODE ContainerNode,
    __in UINT16 Value
    );




NTSTATUS
AddSeqAttribute(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_TREE_ROOT_NODE Tree,
    __in USHORT AttribId,
    __out PSDP_NODE * Seq
    )
{
    NTSTATUS status;
    PSDP_NODE seq; 

	KdPrint(("SCRD! AddSeqAttribute: Enter\n"));

    seq = SdpNodeInterface->SdpCreateNodeSequence(
        POOLTAG_BTHECHOSAMPLE
        );
    
    if (NULL == seq)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

		KdPrint( ("AddSeqAttribute: Creating sequence failed, returning status code %!STATUS!\n", status));
        
        return status;
    }

    status = SdpNodeInterface->SdpAddAttributeToTree(Tree, AttribId, seq, POOLTAG_BTHECHOSAMPLE);
    if(!NT_SUCCESS(status))
    {
		KdPrint( ("AddSeqAttribute: SdpAddAttributeToTree failed, Status code %!STATUS!\n", status));

        //
        // If we failed to add attribute to tree use ExFreePool to free it
        //
        ExFreePool(seq);

        return status;
    }

    *Seq = seq;

    return status;
}



NTSTATUS
AppendSeqNode(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_NODE containerNode,
    __out PSDP_NODE * Node
    )
{
    NTSTATUS status;
    PSDP_NODE node;

	KdPrint(("SCRD! AppendSeqNode: Enter\n"));

    node = SdpNodeInterface->SdpCreateNodeSequence(POOLTAG_BTHECHOSAMPLE);
    if(node == NULL)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

		KdPrint( ("SdpCreateNodeSequence failed, returning status code %!STATUS!\n", status));
        
		return status;
    }
    
    status = SdpNodeInterface->SdpAppendNodeToContainerNode(
        containerNode,
        node
        );

    //
    // SdpAppendNodeToContainerNode doesn't fail due to resource reasons
    // Make sure we passed valid parameters and container node
    //

    ASSERT(NT_SUCCESS(status));

    *Node = node;

    return status;
}




NTSTATUS
AppendNodeUint16(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_NODE ContainerNode,
    __in UINT16 Value
    )
{
    NTSTATUS status;
    PSDP_NODE node;

	KdPrint(("SCRD! AppendNodeUint16: Enter\n"));
    
    node = SdpNodeInterface->SdpCreateNodeUint16(Value, POOLTAG_BTHECHOSAMPLE);
    if(node == NULL)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

		KdPrint( ("Creating NodeUint16 failed, returning status code %!STATUS!\n", status));

        return status;
    }

    status = SdpNodeInterface->SdpAppendNodeToContainerNode(
        ContainerNode,
        node
        );

    if(!NT_SUCCESS(status))
    {
		KdPrint( ("SdpAppendNodeToContainerNode failed, returning status code %!STATUS!\n", status));
        ExFreePool(node);
        node = NULL;
        return status;
    }

    //
    // Make sure we passed valid parameters and container node
    //
    
    ASSERT(NT_SUCCESS(status));

    return status;
}



NTSTATUS
AppendNodeUuid128(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_NODE ContainerNode,
    __in const GUID * Value
    )
{
    NTSTATUS status;
    PSDP_NODE node;

	KdPrint(("SCRD! AppendNodeUuid128: Enter\n"));
    
	// SdpCreateNodeUuid128 is used to allocate and initialize an SDP_NODE structure
	// to a 128-bit UUID with Value uuid parameter
    node = SdpNodeInterface->SdpCreateNodeUuid128(Value, POOLTAG_BTHECHOSAMPLE);
    if(node == NULL)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

		KdPrint( ("Creating NodeUuid128 failed, returning status code %!STATUS!\n", status));

        return status;
    }

    status = SdpNodeInterface->SdpAppendNodeToContainerNode(
        ContainerNode,
        node
        );

    if(!NT_SUCCESS(status))
    {
		KdPrint( ("SdpAppendNodeToContainerNode failed, returning status code %!STATUS!\n", status));
        ExFreePool(node);
        node = NULL;
        return status;
    }

    //
    // Make sure we passed valid parameters and container node
    //
    
    ASSERT(NT_SUCCESS(status));

    return status;
}

NTSTATUS
AppendNodeUuid16(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PSDP_NODE ContainerNode,
    __in UINT16 Value
    )
{
    NTSTATUS status;
    PSDP_NODE node;

	KdPrint(("SCRD! AppendNodeUuid16: Enter\n"));
    
    node = SdpNodeInterface->SdpCreateNodeUuid16(Value, POOLTAG_BTHECHOSAMPLE);
    if(node == NULL)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

		KdPrint( ("Creating NodeUuid16 failed, returning status code %!STATUS!\n", status));

        return status;
    }

    status = SdpNodeInterface->SdpAppendNodeToContainerNode(
        ContainerNode,
        node
        );

    if(!NT_SUCCESS(status))
    {
		KdPrint( ("SdpAppendNodeToContainerNode failed, returning status code %!STATUS!\n", status));
        ExFreePool(node);
        node = NULL;
        return status;
    }

    //
    // Make sure we passed valid parameters and container node
    //
    
    ASSERT(NT_SUCCESS(status));

    return status;
}




NTSTATUS
CreateSdpRecord(
    __in PBTHDDI_SDP_NODE_INTERFACE SdpNodeInterface,
    __in PBTHDDI_SDP_PARSE_INTERFACE SdpParseInterface,
    __in const GUID * ClassId,
    __in LPWSTR Name,
    __in USHORT Psm,
    __out PUCHAR * Stream,
    __out ULONG * Size
    )
/*++
Description:    Create server SDP record
Arguments:
    SdpNodeInterface - Node interface that we obtained from bth stack
    SdpParseInterface - Parse interface that we obtained from bth stack
    ClassId - Service Class ID to publish
    Name - Service name to publish
    Psm - Server PSM
    Stream - receives the sdp record stream
    Size - receives size of sdp record stream
--*/
{
    NTSTATUS status;
    PSDP_TREE_ROOT_NODE tree = NULL;
    PSDP_NODE seqClsIdList, seqProto, seqLang; 
    PSDP_NODE nodeName = NULL, nodeDesc = NULL, nodeProto; 
    UNICODE_STRING unicodeStrName;
    ANSI_STRING ansiStrName;
    PUCHAR stream = NULL;
    ULONG size;
    ULONG_PTR errorByte = 0;

	KdPrint(("SCRD!--> CreateSdpRecord: Enter\n"));

    RtlInitUnicodeString(&unicodeStrName, Name);
    RtlInitAnsiString(&ansiStrName, NULL);

    status = RtlUnicodeStringToAnsiString(&ansiStrName, &unicodeStrName, TRUE);
    if (!NT_SUCCESS(status))
    {
		KdPrint( ("Creating ANSI string for service name failed, Status code %!STATUS!\n", status));

        ansiStrName.Length = 0;
        goto exit;
    }
        
    tree = SdpNodeInterface->SdpCreateNodeTree(
        POOLTAG_BTHECHOSAMPLE
        );

    if (NULL == tree)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

		KdPrint( ("SdpCreateNodeTree failed, returning Status code %!STATUS!\n", status));
        
        goto exit;
    }

    //
    // Add ClassIdList attribute
    //

    status = AddSeqAttribute(SdpNodeInterface, tree, SDP_ATTRIB_CLASS_ID_LIST, &seqClsIdList);
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    status = AppendNodeUuid128(SdpNodeInterface, seqClsIdList, ClassId);
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    //
    // Add protocols
    //

    //
    // L2CAP
    //
    status = AddSeqAttribute(SdpNodeInterface, tree, SDP_ATTRIB_PROTOCOL_DESCRIPTOR_LIST, &seqProto);
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    status = AppendSeqNode(SdpNodeInterface, seqProto, &nodeProto);
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }
    
    status = AppendNodeUuid16(SdpNodeInterface, nodeProto, L2CAP_PROTOCOL_UUID16);
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    status = AppendNodeUint16(SdpNodeInterface, nodeProto, Psm);
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    //
    // SDP
    //

    status = AppendSeqNode(SdpNodeInterface, seqProto, &nodeProto);
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    status = AppendNodeUuid16(SdpNodeInterface, nodeProto, SDP_PROTOCOL_UUID16);            
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    //
    // Add lang attributes
    //
    status = AddSeqAttribute(SdpNodeInterface, tree, SDP_ATTRIB_LANG_BASE_ATTRIB_ID_LIST, &seqLang);
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    status = AppendNodeUint16(SdpNodeInterface, seqLang, 0x656e); //TODO: find constants for these           
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    status = AppendNodeUint16(SdpNodeInterface, seqLang, 0x006A);            
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }
    
    status = AppendNodeUint16(SdpNodeInterface, seqLang, 0x0100);            
    if (!NT_SUCCESS(status))
    {
        goto exit;
    }
    
    //
    // Add service name
    //

    nodeName = SdpNodeInterface->SdpCreateNodeString(
        ansiStrName.Buffer, 
        ansiStrName.Length, 
        POOLTAG_BTHECHOSAMPLE
        );
    if(NULL == nodeName)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

		KdPrint( ("Creating node for service name failed, Status code %!STATUS!\n", status));

        goto exit;
    }
    
    status = SdpNodeInterface->SdpAddAttributeToTree(
        tree, 
        LANG_DEFAULT_ID+STRING_NAME_OFFSET, 
        nodeName, 
        POOLTAG_BTHECHOSAMPLE
        );
    if(!NT_SUCCESS(status))
    {
		KdPrint( ("SdpAddAttributeToTree for service name failed, Status code %!STATUS!\n", status));

        goto exit;
    }

    nodeName = NULL; //transferred owenership to tree
    
    nodeDesc = SdpNodeInterface->SdpCreateNodeString(
        ansiStrName.Buffer, 
        ansiStrName.Length, 
        POOLTAG_BTHECHOSAMPLE
        );
    if(NULL == nodeDesc)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

		KdPrint( ("Creating node for service desc failed, Status code %!STATUS!\n", status));

        goto exit;
    }
    
    status = SdpNodeInterface->SdpAddAttributeToTree(
        tree, 
        LANG_DEFAULT_ID+STRING_DESCRIPTION_OFFSET, 
        nodeDesc, 
        POOLTAG_BTHECHOSAMPLE
        );
    if(!NT_SUCCESS(status))
    {
        KdPrint( ("SdpAddAttributeToTree for service desc failed, Status code %!STATUS!\n", status));
        goto exit;
    }

    nodeDesc = NULL;

    //
    // Create stream from tree
    //

    status = SdpParseInterface->SdpConvertTreeToStream(tree, &stream, &size, POOLTAG_BTHECHOSAMPLE);
    if(!NT_SUCCESS(status))
    {
		KdPrint( ("Failed to get stream from tree for SDP record, Status code %!STATUS!\n", status));
        
        goto exit;
    }

    status = SdpParseInterface->SdpValidateStream(
        stream,
        size,
        &errorByte
        );

    if(!NT_SUCCESS(status))
    {
        KdPrint( ("Validate stream failed for SDP record, first failure at address %p\n", status));
        goto exit;
    }

    *Stream = stream;
    *Size = size;
    
exit:
    if (NULL != tree)
    {
        SdpNodeInterface->SdpFreeTree(tree);
    }

    if (NULL != nodeName)
    {
        //
        // If we failed to add attribute to tree use ExFreePool to free it
        //
        ExFreePool(nodeName);
    }
    
    if (NULL != nodeDesc)
    {
        //
        // If we failed to add attribute to tree use ExFreePool to free it
        //
        ExFreePool(nodeDesc);
    }
    
    if(0 != ansiStrName.Length)
    {
        RtlFreeAnsiString(&ansiStrName);
    }

    if (!NT_SUCCESS(status))
    {
        if (stream != NULL)
        {
            ExFreePoolWithTag(stream, POOLTAG_BTHECHOSAMPLE);
        }
    }
    KdPrint(("SCRD!<-- CreateSdpRecord: Exit\n"));
    return status;
}




VOID
FreeSdpRecord(
    PUCHAR SdpRecord
    )
{
	KdPrint(("SCRD! FreeSdpRecord: Enter\n"));
    ExFreePoolWithTag(SdpRecord, POOLTAG_BTHECHOSAMPLE);
	KdPrint(("SCRD! FreeSdpRecord: Exit\n"));
}