/*
 * Copyright (c) 2010 NTNU, KTH Oleksandr Bodriagov. All rights reserved. :)
 */

/*
 *  JavaCardHOTP.java	 05/03/2010

 *  This class provides secure HMAC-based One Time Password generation
 *  based on the shared secret K and the data received from the phone.
 *  Returned value is a 32-bit byte array.
 */

package totp;


import javacard.framework.APDU;
import javacard.framework.ISO7816;
import javacard.framework.ISOException;
import javacard.framework.Applet;
import javacard.framework.OwnerPIN;
import javacard.framework.JCSystem;
import javacard.framework.Util;
import javacard.security.MessageDigest;





public class JCtotp extends Applet
{
    final static byte VERIFY_PIN = (byte) 0x20; // INS value for ISO 7816-4 VERIFY
    final static byte UPDATE_PIN = (byte) 0x57; // INS value for update PIN request
    final static byte GENERATE_TOTP = (byte) 0x58; // INS value in the generate-request APDUs
    final static byte [] SW_NO_ERROR = {(byte)0x90,(byte)0x00};

    final static byte TryLimit = (byte) 4;
    final static byte MaxPINSize = (byte) 8;
    final static byte secKeyLength = (byte) 32;
    private OwnerPIN appletPIN;
    private byte [] secretKey;
    
    final static byte blockLength = 64;//block lengths in HMAC-SHA1
    private byte [] workbuffer; // buffer for HMAC-SHA method
    private short offset;// offset value from which HOTP truncation is started
    private byte[] hotp; //array for storing final HOTP value     
    private MessageDigest shaDigest;

      
    // Constructor. Only this class's install method should create the applet object.
    protected JCtotp(byte[] installParameters, short offset, byte length)
    {

        byte [] initialPIN = new byte[4];
        short pinOffset = offset;
        
        secretKey = new byte[secKeyLength];        

        //Util.arrayCopy(installParameters, (short)(offset+1), appletAID, (short)0, (short)aidLength);
        Util.arrayCopy(installParameters, (short)(pinOffset), initialPIN,(short)0 , (short)4);
        Util.arrayCopy(installParameters, (short)(pinOffset + 4), secretKey, (short)0, secKeyLength);

        appletPIN = new OwnerPIN(TryLimit, MaxPINSize);
        appletPIN.update(initialPIN, (short) 0,(byte) 4);

        shaDigest = MessageDigest.getInstance(MessageDigest.ALG_SHA,false);
        hotp = JCSystem.makeTransientByteArray((short)20, JCSystem.MEMORY_TYPE_TRANSIENT_RESET);//cleared when applet is deselected
        workbuffer = JCSystem.makeTransientByteArray((short)(blockLength + MessageDigest.LENGTH_SHA), JCSystem.MEMORY_TYPE_TRANSIENT_RESET);

        register();
    }


    // Installs applet. installParameters: contains installation parameters ( ,secret key, )
    public static void install(byte[] bArray, short bOffset, byte bLength)
    {
        new JCtotp(bArray, bOffset, bLength);
    }


    // Processes incoming APDUs
    public void process(APDU apdu)
    {
        byte[] buffer = apdu.getBuffer(); //gets a pointer to the APDU buffer

	    // check if it is SELECT APDU command
        if ( buffer[ISO7816.OFFSET_CLA] == 0 && buffer[ISO7816.OFFSET_INS] == (byte) (0xA4) )
        return;

        switch (buffer[ISO7816.OFFSET_INS])
        {
             case VERIFY_PIN:
                    verifyPIN(apdu);
                    break;
             case GENERATE_TOTP:
                    generateTOTP(apdu);
                    break;
             case UPDATE_PIN:
                    updatePIN(apdu);
                    break;
             default:
                    ISOException.throwIt(ISO7816.SW_INS_NOT_SUPPORTED);
        }
    }



    // user authentication via PIN check
    private void verifyPIN(APDU apdu)
    {
        byte[] buffer = apdu.getBuffer();
        byte dataLength = buffer[ISO7816.OFFSET_LC]; // number of bytes in the data field of APDU
        if (dataLength > 8 || dataLength < 1)
            ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
        apdu.setIncomingAndReceive(); //data is read into apdu buffer

        if (appletPIN.check(buffer, ISO7816.OFFSET_CDATA, dataLength ) != true)
        	ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
        else
        {
        	apdu.setOutgoing();
            apdu.setOutgoingLength((short) SW_NO_ERROR.length);
            apdu.sendBytesLong(SW_NO_ERROR, (short) 0, (short)SW_NO_ERROR.length);
        }

    }
    
    
    private void updatePIN(APDU apdu)
    {
        if(appletPIN.isValidated() != true)
            ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
        appletPIN.reset();//resets the validated flag and resets the PIN try counter to the value
        
        byte[] buffer = apdu.getBuffer();
        byte dataLength = buffer[ISO7816.OFFSET_LC]; // number of bytes in the data field of APDU
        apdu.setIncomingAndReceive(); //data is read into apdu buffer
        
        if (dataLength > 8 || dataLength < 4)
            ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
        appletPIN.update(buffer, ISO7816.OFFSET_CDATA, dataLength);
        appletPIN.resetAndUnblock();
        
    	apdu.setOutgoing();
        apdu.setOutgoingLength((short) SW_NO_ERROR.length);
        apdu.sendBytesLong(SW_NO_ERROR, (short) 0, (short)SW_NO_ERROR.length);
	}



    private void generateTOTP(APDU apdu)
    {
        if(appletPIN.isValidated() != true)
            ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
        appletPIN.reset();//resets the validated flag and resets the PIN try counter to the value

        byte[] buffer = apdu.getBuffer();
        byte dataLength = buffer[ISO7816.OFFSET_LC]; // number of bytes in the data field of APDU
        if (dataLength > 8)
            ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
        apdu.setIncomingAndReceive(); //data is read into apdu buffer

        limitedHmacSha(hotp, secretKey, buffer, (short)ISO7816.OFFSET_CDATA,(short)dataLength);

        // Dynamic truncation according to RFC 4226
        offset = (short)(hotp[MessageDigest.LENGTH_SHA - 1] & 0xf); //take only 4 low-order bits

        apdu.setOutgoing();
        apdu.setOutgoingLength((short) 4);
        apdu.sendBytesLong(hotp, (short) offset, (short) 4);
    }



    //calculates HMAC-SHA1 according to RFC 2104 for data blocks <= 20 bytes
    private void limitedHmacSha (byte [] mac, byte [] key,  byte [] data, short dOffset, short datalength)
    {
        byte i = 0;

        for (i = 0; i < blockLength; i++)
           workbuffer[i] = (byte)0x00;

        //hash the key if it is longer than 64 bytes and write result to workbuffer
        if (key.length > blockLength)
           shaDigest.doFinal(key, (short) 0, (short)key.length, workbuffer, (short) 0);
        else
           for (i = 0; i < key.length; i++)
           workbuffer[i] = key[i];

        //K XOR ipad
        for (i = 0; i < blockLength; i++)
            workbuffer [i] ^= (byte) 0x36;

        //append the stream of data 'text' to the (K XOR ipad)
        for (i = 0; i < datalength; i++)
            workbuffer[(short)(blockLength + i)] = data [(short)(dOffset + i)];

         //creates H(K XOR ipad, text) and appends to the workbuffer after the space for (K XOR opad)
        shaDigest.doFinal(workbuffer, (short) 0, (short)(blockLength + datalength), workbuffer, (short) blockLength);

        //calculates (K XOR opad) and puts it in the buffer before H(K XOR ipad, text)
        for (i = 0; i < blockLength; i++)
        {
            workbuffer [i] ^= (byte) 0x36;//take away previous XOR
            workbuffer [i] ^= 0x5C;
        }

        // calculates H(K XOR opad, H(K XOR ipad, text)) and writes it to mac
        shaDigest.doFinal (workbuffer, (short) 0, (short) (blockLength + MessageDigest.LENGTH_SHA), mac, (short) 0);
    }



}//end class JCtotp