

/*
 * Copyright (c) 2010 NTNU/KTH; author: Oleksandr Bodriagov. All rights reserved. :)

 * Class CardCommunicator: handles communication with the Java Card applet
 */
package totp;

import java.util.*;
import javax.microedition.apdu.APDUConnection;


public class CardCommunicator
{
    private APDUConnection cardConnection;

    final static byte [] SW_NO_ERROR = {(byte)0x90,(byte)0x00};
    final static byte INS_VERIFY_PIN = (byte) 0x20; // INS value for ISO 7816-4 VERIFY
    final static byte INS_CHANGE_PIN = (byte) 0x57; // INS value for the Update PIN command
    final static byte INS_GENERATE_TOTP = (byte) 0x58; // INS value for the Generate TOTP command

    public CardCommunicator()
    {
        cardConnection = null;
    }

    public void setConnection (APDUConnection connection)
    {
        cardConnection = connection;
    }


    //sends data to the card for TOTP generation
    public String generateTOTP(int numbDigits)
    {
        if (cardConnection == null)
            return null;
        try
        {
            byte[] totpvalue = cardConnection.exchangeAPDU(makeTimeEnvelope());

            if (totpvalue.length > 2)//received some data, not the error notification
            {
                int result = 0;

                // byte [] ---> int conversion
                result = (totpvalue[0] & 0x7f) << 24
                       | (totpvalue[1] & 0xff) << 16
                       | (totpvalue[2] & 0xff) << 8
                       | (totpvalue[3] & 0xff);
                result = result%1000000;
                String totp = Integer.toString(result);
                if (totp.length() < numbDigits)
                     totp = "0" + totp;
                return totp;            
            }
            else return null;
        }
        catch (Exception e)
        {
            return null;
        }
    }


    //sends PIN to the Java Card applet for verification
    public boolean verifyPIN (String pinCode)
    {
        if (cardConnection == null)
            return false;
        try
        {
            byte[] response = cardConnection.exchangeAPDU(makePinApdu(INS_VERIFY_PIN, pinCode));

            if (response[0] == SW_NO_ERROR[0] && response[1] == SW_NO_ERROR[1])
                return true;
            else
                return false;
        }
        catch (Exception e)
        {
            return false;
        }

    }

    //sends new String PIN to the Java Card applet for verification
    public boolean updatePIN (String newPIN)
    {
        if (cardConnection == null)
            return false;

        try
        {
            byte[] response = cardConnection.exchangeAPDU(makePinApdu(INS_CHANGE_PIN, newPIN ));

            if (response[0] == SW_NO_ERROR[0] && response[1] == SW_NO_ERROR[1])
                return true;
            else
                return false;
        }
        catch (Exception e)
        {
            return false;
        }
    }


    //creates envelope that contains entered PIN from a String
    private byte[] makePinApdu(byte command, String pin)
    {
        if (pin == null)
            return null;

        int pinlength = pin.length();
        byte [] buffer = pin.getBytes();
        byte [] pinEnvelope = new byte [6 + pinlength];
        pinEnvelope [0] = (byte)0x80;//APDU CLA

        switch(command)
        {
            case INS_VERIFY_PIN:
                pinEnvelope [1] = (byte)INS_VERIFY_PIN;//APDU INS
                break;
            case INS_CHANGE_PIN:
                pinEnvelope [1] = (byte)INS_CHANGE_PIN;//APDU INS
                break;
            default:
                return null;
        }

        pinEnvelope [2] = (byte)0;//APDU P1
        pinEnvelope [3] = (byte)0;//APDU P2
        pinEnvelope [4] = (byte)pinlength;//APDU Lc
        for (int i=0; i < pinlength; i++)
            pinEnvelope [5+i] = buffer[i];

        pinEnvelope [5 + pinlength] = (byte)0;
        return pinEnvelope;
     }


     //creates APDU envelope that contains current time (from 1970)
     private byte [] makeTimeEnvelope()
     {
         Date clock = new Date();
         long time = clock.getTime()/1000;//time - number of seconds from 1970

         // long --> byte [8] convertion
	 byte[] buffer = new byte[8];
	 for(int i = 0;i < 8; i++)
         {
	     buffer[i] = (byte)(time & 0xff);
             time = time >>> 8;
         }
         byte [] timeEnvelope = new byte [6 + buffer.length];
         timeEnvelope [0] = (byte)0x80;//APDU CLA
         timeEnvelope [1] = INS_GENERATE_TOTP;//APDU INS
         timeEnvelope [2] = (byte)0;//APDU P1
         timeEnvelope [3] = (byte)0;//APDU P2
         timeEnvelope [4] = (byte)buffer.length;//APDU LC - number of bytes present in the data field

         for (int i=0; i < buffer.length; i++)
         timeEnvelope [5+i] = buffer[i];//APDU Data

         timeEnvelope [5 + buffer.length] = (byte)0;//APDU Le

         return timeEnvelope;
     }


}
